#################################################################################
# DISPATCHES was produced under the DOE Design Integration and Synthesis Platform
# to Advance Tightly Coupled Hybrid Energy Systems program (DISPATCHES), and is
# copyright (c) 2020-2023 by the software owners: The Regents of the University
# of California, through Lawrence Berkeley National Laboratory, National
# Technology & Engineering Solutions of Sandia, LLC, Alliance for Sustainable
# Energy, LLC, Battelle Energy Alliance, LLC, University of Notre Dame du Lac, et
# al. All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. Both files are also available online at the URL:
# "https://github.com/gmlc-dispatches/dispatches".
#################################################################################


# Import Pyomo libraries
from pyomo.common.config import ConfigBlock, ConfigValue, In
from pyomo.environ import Var, NonNegativeReals

# Import IDAES cores
from idaes.core import declare_process_block_class, UnitModelBlockData
from idaes.core.util.config import is_physical_parameter_block
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.core.solvers.get_solver import get_solver
import idaes.logger as idaeslog

__author__ = "Radhakrishna Tumbalam Gooty"


# Set up logger
_log = idaeslog.getLogger(__name__)


@declare_process_block_class("SimpleHydrogenTank")
class SimpleHydrogenTankData(UnitModelBlockData):
    """
    Simplified tank model where energy balance is ignored
    """

    CONFIG = ConfigBlock()
    CONFIG.declare(
        "dynamic",
        ConfigValue(
            domain=In([False]),
            default=False,
            description="Dynamic model flag - must be False",
            doc="""Translator blocks are always steady-state.""",
        ),
    )
    CONFIG.declare("has_holdup", ConfigValue(
        default=False,
        domain=In([False]),
        description="Holdup construction flag - must be False",
        doc="""Indicates whether holdup terms should be constructed or not.
    **default** - False. Hydrogen tank model uses custom equations for holdup.""")
    )
    CONFIG.declare(
        "property_package",
        ConfigValue(
            default=None,
            domain=is_physical_parameter_block,
            description="Property package to use for incoming stream",
            doc="""Property parameter object used to define property
calculations for the incoming stream,
**default** - None.
**Valid values:** {
**PhysicalParameterObject** - a PhysicalParameterBlock object.}""",
        ),
    )
    CONFIG.declare(
        "property_package_args",
        ConfigBlock(
            implicit=True,
            description="Arguments to use for constructing property package "
            "of the incoming stream",
            doc="""A ConfigBlock with arguments to be passed to the property
block associated with the incoming stream,
**default** - None.
**Valid values:** {
see property package for documentation.}""",
        ),
    )

    def build(self):
        """
        Begin building model.
        Args:
            None
        Returns:
            None
        """
        # Call UnitModel.build to setup dynamics
        super(SimpleHydrogenTankData, self).build()

        # Add State Blocks
        self.properties_in = self.config.property_package.build_state_block(
            self.flowsheet().time,
            doc="Material properties in incoming stream",
            defined_state=True,
            has_phase_equilibrium=False,
            **self.config.property_package_args,
        )

        self.properties_out_pipeline = self.config.property_package.build_state_block(
            self.flowsheet().time,
            doc="Material properties in outlet stream to pipeline",
            defined_state=True,
            has_phase_equilibrium=False,
            **self.config.property_package_args,
        )

        self.properties_out_turbine = self.config.property_package.build_state_block(
            self.flowsheet().time,
            doc="Material properties in outlet stream to turbine",
            defined_state=True,
            has_phase_equilibrium=False,
            **self.config.property_package_args,
        )

        # Add outlet port
        self.add_port(name="inlet",
                      block=self.properties_in,
                      doc="Inlet Port")
        self.add_port(name="outlet_to_turbine",
                      block=self.properties_out_turbine,
                      doc="Outlet Port to Hydrogen Turbine")
        self.add_port(name="outlet_to_pipeline",
                      block=self.properties_out_pipeline,
                      doc="Outlet Port to Pipeline")

        # Keep the temperature and pressure the same for inlets and outlets
        @self.Constraint(self.flowsheet().time)
        def eq_temperature_1(blk, t):
            return (
                blk.properties_in[t].temperature ==
                blk.properties_out_turbine[t].temperature
            )

        @self.Constraint(self.flowsheet().time)
        def eq_temperature_2(blk, t):
            return (
                blk.properties_in[t].temperature ==
                blk.properties_out_pipeline[t].temperature
            )

        @self.Constraint(self.flowsheet().time)
        def eq_pressure_1(blk, t):
            return (
                blk.properties_in[t].pressure ==
                blk.properties_out_turbine[t].pressure
            )

        @self.Constraint(self.flowsheet().time)
        def eq_pressure_2(blk, t):
            return (
                blk.properties_in[t].pressure ==
                blk.properties_out_pipeline[t].pressure
            )

        # Get units from property package
        units = self.config.property_package.\
            get_metadata().get_derived_units

        # Define holdup variable
        self.tank_holdup_previous = Var(self.flowsheet().time,
                                        within=NonNegativeReals,
                                        units=units("amount"))
        self.tank_holdup = Var(self.flowsheet().time,
                               within=NonNegativeReals,
                               units=units("amount"))
        self.dt = Var(self.flowsheet().time,
                      initialize=3600,
                      within=NonNegativeReals,
                      units=units("time"))

        # Material balance
        @self.Constraint(self.flowsheet().time)
        def tank_material_balance(blk, t):
            return (
                blk.tank_holdup[t] - blk.tank_holdup_previous[t] ==
                blk.dt[t] * (+ blk.properties_in[t].flow_mol
                             - blk.properties_out_pipeline[t].flow_mol
                             - blk.properties_out_turbine[t].flow_mol)
            )

    def initialize_build(
        blk,
        state_args_in=None,
        state_args_out=None,
        outlvl=idaeslog.NOTSET,
        solver=None,
        optarg=None,
    ):
        """
        This method calls the initialization method of the state blocks.
        Keyword Arguments:
            state_args_in : a dict of arguments to be passed to the inlet
                            property package (to provide an initial state for
                            initialization (see documentation of the specific
                            property package) (default = None).
            state_args_out : a dict of arguments to be passed to the outlet
                             property package (to provide an initial state for
                             initialization (see documentation of the specific
                             property package) (default = None).
            outlvl : sets output level of initialization routine
            optarg : solver options dictionary object (default=None, use
                     default solver options)
            solver : str indicating which solver to use during
                     initialization (default = None, use default solver)
        Returns:
            None
        """
        init_log = idaeslog.getInitLogger(blk.name, outlvl, tag="unit")

        # Create solver
        opt = get_solver(solver, optarg)

        # ---------------------------------------------------------------------
        # Initialize state block
        flags = blk.properties_in.initialize(
            outlvl=outlvl,
            optarg=optarg,
            solver=solver,
            state_args=state_args_in,
            hold_state=True,
        )

        blk.properties_out_turbine.initialize(
            outlvl=outlvl,
            optarg=optarg,
            solver=solver,
            state_args=state_args_out,
        )

        blk.properties_out_pipeline.initialize(
            outlvl=outlvl,
            optarg=optarg,
            solver=solver,
            state_args=state_args_out,
        )

        if degrees_of_freedom(blk) != 0:
            raise AssertionError(f"Initialization of {blk.name} is unsuccessful. "
                                 f"Degrees of freedom was not zero. Please provide "
                                 f"sufficient number of constraints linking the state "
                                 f"variables between the two state blocks.")

        with idaeslog.solver_log(init_log, idaeslog.DEBUG) as slc:
            res = opt.solve(blk, tee=slc.tee)

        init_log.info("Initialization Complete {}."
                      .format(idaeslog.condition(res)))

        blk.properties_in.release_state(flags=flags, outlvl=outlvl)



#################################################################################
# DISPATCHES was produced under the DOE Design Integration and Synthesis Platform
# to Advance Tightly Coupled Hybrid Energy Systems program (DISPATCHES), and is
# copyright (c) 2020-2023 by the software owners: The Regents of the University
# of California, through Lawrence Berkeley National Laboratory, National
# Technology & Engineering Solutions of Sandia, LLC, Alliance for Sustainable
# Energy, LLC, Battelle Energy Alliance, LLC, University of Notre Dame du Lac, et
# al. All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. Both files are also available online at the URL:
# "https://github.com/gmlc-dispatches/dispatches".
#################################################################################
"""
Turbo-Generator Set for a Hydrogen turbine.

Compressor -> Stoichiometric Reactor -> Turbine
Author: Konor Frick
Date: April 2, 2021
Notes: it is noted that in this example the hydrogen is compressed along
with the air in the compressor as opposed to having a separate
fuel injection system. Noting this is a simplified version of the H2 turbine.
"""
from pyomo.environ import Constraint, Var, TransformationFactory
from pyomo.common.config import ConfigBlock, ConfigValue, In
from pyomo.network import Arc

from idaes.models.unit_models import Compressor, \
    StoichiometricReactor, Turbine

import idaes.logger as idaeslog

from idaes.core.util.config import is_physical_parameter_block, \
    is_reaction_parameter_block
from idaes.core.util.initialization import propagate_state
from idaes.core.solvers.get_solver import get_solver
from idaes.core import declare_process_block_class, UnitModelBlockData, \
    useDefault


@declare_process_block_class(
    "HydrogenTurbine",
    doc="Simple 0D hydrogen turbine unit model")
class HydrogenTurbineData(UnitModelBlockData):
    """
    Simple 0D hydrogen unit model that consists of three units:
    Compressor --> Stoichiometric Reaactor --> Turbine.
    """
    CONFIG = ConfigBlock()
    CONFIG.declare("dynamic", ConfigValue(
        domain=In([False]),
        default=False,
        description="Dynamic model flag - must be False",
        doc="""HydrogenTurbine does not support dynamic models,
        thus this must be False."""))
    CONFIG.declare("has_holdup", ConfigValue(
        default=False,
        domain=In([False]),
        description="Holdup construction flag",
        doc="""HydrogenTurbine does not support dynamic models,
        thus this must be False."""))
    CONFIG.declare("property_package", ConfigValue(
        default=useDefault,
        domain=is_physical_parameter_block,
        description="Property package to use for control volume",
        doc="""Property parameter object used to define property calculations,
**default** - useDefault.
**Valid values:** {
**useDefault** - use default package from parent model or flowsheet,
**PropertyParameterObject** - a PropertyParameterBlock object.}"""))
    CONFIG.declare("property_package_args", ConfigBlock(
        implicit=True,
        description="Arguments to use for constructing property packages",
        doc="""A ConfigBlock with arguments to be passed to a property block(s)
and used when constructing these,
**default** - None.
**Valid values:** {
see property package for documentation.}"""))
    CONFIG.declare("reaction_package", ConfigValue(
        default=None,
        domain=is_reaction_parameter_block,
        description="Reaction package to use for control volume",
        doc="""Reaction parameter object used to define reaction calculations,
**default** - None.
**Valid values:** {
**None** - no reaction package,
**ReactionParameterBlock** - a ReactionParameterBlock object.}"""))
    CONFIG.declare("reaction_package_args", ConfigBlock(
        implicit=True,
        description="Arguments to use for constructing reaction packages",
        doc="""A ConfigBlock with arguments to be passed to a reaction block(s)
and used when constructing these,
**default** - None.
**Valid values:** {
see reaction package for documentation.}"""))

    def build(self):

        # Call UnitModel.build to setup dynamics
        super(HydrogenTurbineData, self).build()

        self.compressor = Compressor(property_package=self.config.property_package)

        self.stoic_reactor = StoichiometricReactor(
            property_package=self.config.property_package,
            reaction_package=self.config.reaction_package,
            has_heat_of_reaction=True,
            has_heat_transfer=False,
            has_pressure_change=False,
        )

        self.turbine = Turbine(property_package=self.config.property_package)

        # Declare var for reactor conversion
        self.stoic_reactor.conversion = Var(initialize=0.75, bounds=(0, 1))

        stoic_reactor_in = self.stoic_reactor.control_volume.properties_in[0.0]
        stoic_reactor_out = self.stoic_reactor.control_volume.properties_out[0.0]

        self.stoic_reactor.conv_constraint = Constraint(
            expr=self.stoic_reactor.conversion *
            stoic_reactor_in.flow_mol_comp["hydrogen"] ==
            (stoic_reactor_in.flow_mol_comp["hydrogen"] -
             stoic_reactor_out.flow_mol_comp["hydrogen"]))

        # Connect arcs
        self.comp_to_reactor = Arc(
            source=self.compressor.outlet,
            destination=self.stoic_reactor.inlet)
        self.reactor_to_turbine = Arc(
            source=self.stoic_reactor.outlet,
            destination=self.turbine.inlet)

        # Expression for net mechanical work
        @self.Expression(self.parent_block().time)
        def work_mechanical(blk, t):
            return blk.compressor.work_mechanical[t] + blk.turbine.work_mechanical[t]

        TransformationFactory("network.expand_arcs").apply_to(self)

    def initialize_build(self, state_args=None,
                   solver=None, optarg=None, outlvl=idaeslog.NOTSET):

        init_log = idaeslog.getInitLogger(self.name, outlvl, tag="unit")
        init_log.info_low("Starting initialization...")

        self.compressor.initialize(state_args=state_args, outlvl=outlvl)
        propagate_state(self.comp_to_reactor)

        self.stoic_reactor.initialize(outlvl=outlvl)
        propagate_state(self.reactor_to_turbine)

        self.turbine.initialize(outlvl=outlvl)
        init_log.info_low("Initialization complete")

    def report(self, **kwargs):
        self.compressor.report()
        self.stoic_reactor.report()
        self.turbine.report()

    @property
    def inlet(self):
        return self.compressor.inlet

    @property
    def outlet(self):
        return self.turbine.outlet



#################################################################################
# DISPATCHES was produced under the DOE Design Integration and Synthesis Platform
# to Advance Tightly Coupled Hybrid Energy Systems program (DISPATCHES), and is
# copyright (c) 2020-2023 by the software owners: The Regents of the University
# of California, through Lawrence Berkeley National Laboratory, National
# Technology & Engineering Solutions of Sandia, LLC, Alliance for Sustainable
# Energy, LLC, Battelle Energy Alliance, LLC, University of Notre Dame du Lac, et
# al. All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. Both files are also available online at the URL:
# "https://github.com/gmlc-dispatches/dispatches".
#################################################################################

"""Test for ultra supercritical power plant integrated with discharge
storage system

"""

__author__ = "Soraya Rawlings"

import pytest

from pyomo.environ import TerminationCondition, value, SolverFactory
from pyomo.contrib.fbbt.fbbt import  _prop_bnds_root_to_leaf_map
from pyomo.core.expr.numeric_expr import ExternalFunctionExpression

from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.core.solvers import get_solver

from dispatches.case_studies.fossil_case.ultra_supercritical_plant import (
    ultra_supercritical_powerplant as usc)

from dispatches.case_studies.fossil_case.ultra_supercritical_plant.storage \
     import discharge_design_ultra_supercritical_power_plant as discharge_usc


optarg = {"max_iter": 150,
          "tol": 1e-8,
          "halt_on_ampl_error": "yes"}
solver = get_solver('ipopt', optarg)
add_efficiency = True
power_max = 436
heat_duty = 148.5


@pytest.fixture(scope="module")
def model():

    # Build ultra-supercritical plant base model
    m_usc = usc.build_plant_model()

    # Initialize ultra-supercritical plant base model
    usc.initialize(m_usc)

    m = discharge_usc.create_discharge_model(m_usc,
                                             add_efficiency=add_efficiency,
                                             power_max=power_max)

    # Give all the required inputs to the model
    discharge_usc.set_model_input(m)

    # Add disjunction
    discharge_usc.add_disjunction(m)

    # Add scaling factor
    discharge_usc.set_scaling_factors(m)

    return m


@pytest.mark.integration
def test_main_function():

    # Build ultra-supercritical plant base model
    m_usc = usc.build_plant_model()

    # Initialize ultra-supercritical plant base model
    usc.initialize(m_usc)

    # Build discharge model
    m = discharge_usc.main(m_usc, solver=solver, optarg=optarg)

    discharge_usc.model_analysis(m, heat_duty=heat_duty)

    # Solve model using GDPopt
    results = discharge_usc.run_gdp(m)

    # Print results
    discharge_usc.print_results(m, results)


@pytest.mark.integration
def test_initialize(model):
    # Check that the discharge model is initialized properly
    # The model at this point should have 18 degrees of freedom because of
    # the model is not transformed and have disconnected arcs
    discharge_usc.initialize(model, solver=solver, optarg=optarg)
    assert degrees_of_freedom(model) == 18


@pytest.mark.integration
def test_costing(model):
    # Check that the cost correlations in discharge model are initialized
    # properly
    # The model at this point should have 18 degrees of freedom because of
    # the model is not transformed and have disconnected arcs
    discharge_usc.build_costing(model, solver=solver, optarg=optarg)
    assert degrees_of_freedom(model) == 18


@pytest.mark.integration
def test_usc_discharge_model(model):

    # Unfix disjuncts
    discharge_usc.unfix_disjuncts_post_initialization(model)

    # Add missing functions to complete the discharge model (add bounds
    # and disjunctions)
    discharge_usc.add_bounds(model, power_max=power_max)

    # Add design optimization problem
    discharge_usc.model_analysis(model, heat_duty=heat_duty)

    opt = SolverFactory('gdpopt')
    _prop_bnds_root_to_leaf_map[ExternalFunctionExpression] = lambda x, y, z: None

    result = opt.solve(
        model,
        tee=False,
        algorithm='RIC',
        mip_solver='cbc',
        nlp_solver='ipopt',
        init_algorithm="no_init",
        subproblem_presolve=False,
        nlp_solver_args=dict(options={"max_iter": 150})
    )

    assert result.solver.termination_condition == TerminationCondition.optimal
    assert value(
        model.fs.discharge.condpump_source_disjunct.indicator_var) == True
    assert value(model.fs.discharge.hxd.area) == pytest.approx(1912.2,
                                                               abs=1e-1)
# @pytest.mark.integration
# def test_unit_consistency(model):
#     assert_units_consistent(model)



#################################################################################
# DISPATCHES was produced under the DOE Design Integration and Synthesis Platform
# to Advance Tightly Coupled Hybrid Energy Systems program (DISPATCHES), and is
# copyright (c) 2020-2023 by the software owners: The Regents of the University
# of California, through Lawrence Berkeley National Laboratory, National
# Technology & Engineering Solutions of Sandia, LLC, Alliance for Sustainable
# Energy, LLC, Battelle Energy Alliance, LLC, University of Notre Dame du Lac, et
# al. All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. Both files are also available online at the URL:
# "https://github.com/gmlc-dispatches/dispatches".
#################################################################################

__author__ = "Gabriel J. Soto"

# This file contains utility functions for constructing and using TEAL cashflows and metrics.
# First, we add TEAL to the current Python path. Note that DISPATCHES, TEAL, and RAVEN are all
#   assumed to be subdirectories within the same directory.
import numpy as np
import operator
from TEAL.src import CashFlows
from TEAL.src import main as RunCashFlow
from TEAL.src.Amortization import MACRS

###################
def checkAmortization(compLife, amortYears=None):
  """
    Check proposed amortization schedule against intended component life.
    If amortization schedule not provided, calculates an appropriate
    one that is less than component life.

    @ In, compLife, CashFlow, component lifetime
    @ In, amortYears, float or int, intended amortization years (defaults to None)
    @ Out, amortYears, float or int, corrected amortization years
  """
  MACRS_yrs = np.array(list(MACRS.keys())) # available amortization years
  amortIsCorrect = bool(amortYears is not None and compLife > amortYears) # check if recalc is needed

  # amortization years longer than intended project life, must recalculate
  if not amortIsCorrect:
    assert isinstance(amortYears, (float, int))
    amortYears = MACRS_yrs[compLife > MACRS_yrs].max() # largest value less than project life
    print("Proposed amortization schedule cannot be longer than intended project life.")
    print(f"Returning a shortened schedule: {amortYears} yrs")

  return amortYears

def build_econ_settings(cfs, life=5, dr=0.1, tax=0.21, infl=0.02184, metrics=None):
  """
    Constructs global settings for economic run.
    Repurposed from TEAL/tests/PyomoTest.py

    @ In, cfs, CashFlow, cash flow components
    @ In, life, float, life time of the years to evaluate
    @ In, dr, float, discount rate
    @ In, tax, float, the amount of tax ratio to apply
    @ In, infl, float, the amount of inflation ratio to apply
    @ In, metrics, list, economic metrics to calculate with cashflows
    @ Out, settings, CashFlow.GlobalSettings, settings
  """
  available_metrics = ['NPV',] # TODO: add IRR, PI

  # check against possible economic metrics supported by TEAL that provide Pyomo expressions
  if metrics is not None:
    residual = set(metrics) - set(available_metrics) # extra metrics that are not supported
    if len(residual) > 0:
      raise Exception(f"Requested metrics not in supported list: {available_metrics}")
  else:
    metrics = available_metrics

  active = []
  for comp_name, cf_list in cfs.items():
    for cf in cf_list:
      active.append(f'{comp_name}|{cf}')
  assert 'NPV' in metrics
  params = {'DiscountRate': dr,
            'tax': tax,
            'inflation': infl,
            'ProjectTime': life,
            'Indicator': {'name': metrics, # TODO: check IRR, PI
                          'active': active},
            'Output' : True,
           }
  settings = CashFlows.GlobalSettings()
  settings.setParams(params)
  setattr(settings, "_verbosity", 0)
  return settings

def build_TEAL_Component(name, comp, mdl, scenario=None, scenario_ind=None):
  """
    Constructs TEAL component which holds all desired cash flows (capex, etc.)
    for a given Plant component (PEM, H2 Tank, etc.)

    @ In, name, str, name of Plant Component
    @ In, comp, dict, dictionary of Plant component cash flows
    @ In, mdl, Pyomo model, multiperiod Pyomo model
    @ In, scenario, Pyomo model, submodel of mdl pertaining to a scenario (defaults to None)
    @ In, scenario_ind, int, index number of scenario (defaults to None)
    @ Out, tealComp, CashFlow.Component, object with all cash flows for given plant component
  """
  # if scenario is None, we are doing LMP Deterministic (only 1 scenario)
  # otherwise, mdl is the main Pyomo model and scenario is the submodel for the scenario within mdl
  scenario = mdl if scenario is None else scenario
  life = comp['Lifetime']
  tealComp = CashFlows.Component()
  tealComp.setParams({'name': name,
                      'Life_time': life})
  cashFlows = []
  for cfName, cfDict in comp.items():
    if cfName == 'Capex':
      alpha, driver = getCapexVarFromModel(cfDict, scenario) # get Pyomo expressions
      capex = createCapex(alpha, driver, life) # create actual TEAL cash flow
      cashFlows.append(capex)

      if 'Amortization' in cfDict.keys():
        # check desired time < comp lifetime
        amort = checkAmortization( life, cfDict['Amortization'] )
        capex.setAmortization('MACRS', amort) # calculate schedule
        amorts = getattr(tealComp, '_createDepreciation')(capex) # create actual TEAL cash flow
        cashFlows.extend(amorts)

    elif cfName == 'FixedOM':
      alpha, driver = getCapexVarFromModel(cfDict, scenario)
      fixedOM = createRecurringYearly(alpha, driver, mdl.project_years)
      cashFlows.append(fixedOM)

    elif cfName == 'Hourly':
      alpha, driver = getDispatchVarFromModel(cfDict, mdl, scenario, scenario_ind)
      hourly = createRecurringHourly(alpha, driver, mdl.plant_life)
      cashFlows.append(hourly)

  tealComp.addCashflows(cashFlows)
  return tealComp

def calculate_TEAL_metrics(tealSettings, tealComponentList):
  """
    Calculates desired TEAL metrics (NPV, PI, etc.) given a list of
    plant components and associated cash flows.

    @ In, tealSettings, CashFlow.GlobalSetting, global economic settings
    @ In, tealComponentList, list, list of component cashflows
    @ Out, metrics, dict, dictionary of Pyomo expressions for requested economic metrics
  """
  metrics = RunCashFlow.run(tealSettings, tealComponentList, {}, pyomoVar=True)

  return metrics

####################################################################
# Methods to get Pyomo Expressions and convert to Cash Flow drivers
####################################################################

def getCapexVarFromModel(cfDict, scenario):
  """
    Get Capex parameters from dictionary and convert into a Pyomo expression.
    More specifically, a string is extracted from the dictionary which specifies the
    IDAES Pyomo expression meant to represent the Capex cashflow driver.

    @ In, cfDict, dict, cash flow dictionary
    @ In, scenario, Pyomo model, submodel of mdl pertaining to a scenario (defaults to None)
    @ Out, alpha, float, conversion of driver units to monetary value
    @ Out, driver, Pyomo Expression, cash flow driver
  """
  alpha = cfDict['Value']
  mults = cfDict['Multiplier']
  exprs = cfDict['Expressions']
  assert( len(mults)==len(exprs) )  # NOTE: Multiplier same length as Driver

  # extraction of attribute looks like: model.expression[i]
  pyomoExpr = [operator.attrgetter(exprs[i])(scenario) for i in range(len(exprs))]
  driver = [m*pexp for m, pexp in zip(mults, pyomoExpr)]
  driver = driver[0]
  return alpha, driver

def getDispatchVarFromModel(cfDict, mdl, scenario, scenario_ind=None):
  """
    Get Dispatch parameters from dictionary and convert into a Pyomo expression.
    More specifically, a string is extracted from the dictionary which specifies the
    IDAES Pyomo expression meant to represent the Hourly Dispatch cash flow driver.

    @ In, cfDict, dict, cash flow dictionary
    @ In, mdl, Pyomo model, multiperiod Pyomo model
    @ In, scenario, Pyomo model, submodel of mdl pertaining to a scenario (defaults to None)
    @ In, scenario_ind, int, index number of scenario (defaults to None)
    @ Out, alpha, float, conversion of driver units to monetary value
    @ Out, dispatch_array, numpy array, array of hourly cashflows (Pyomo expressions)
  """
  alpha = cfDict['Value']
  mults = cfDict['Multiplier']
  exprs = cfDict['Expressions']
  assert( len(mults)==len(exprs) )  # NOTE: Multiplier same length as Driver

  # time indeces for HERON/TEAL
  n_hours = len(mdl.set_time)
  n_days  = len(mdl.set_days)
  n_years = len(mdl.set_years)
  n_hours_per_year = n_hours * n_days # sometimes number of days refers to clusters < 365

  n_projLife = mdl.plant_life + 1
  fullYearsArray = np.hstack([0, mdl.project_years]) # looks like [0, 2022, 2022, 2022, 2022, ... 2032, ... ]

  # template array for holding dispatch Pyomo expressions/objects
  dispatch_array = np.zeros((n_projLife, n_hours_per_year), dtype=object)

  # time indeces for DISPATCHES, as array of tuples
  indeces    = np.array([tuple(i) for i in mdl.set_period], dtype="i,i,i")
  time_shape = (n_years, n_hours_per_year) # reshaping the tuples array to match HERON dispatch
  indeces    = indeces.reshape(time_shape)

  is_stochastic = getattr(mdl, '_stochastic_model')
  weights_days = mdl.weights_days[scenario_ind] if is_stochastic else mdl.weights_days

  # currently, taking this to mean that we are using the LMP signal...
  # TODO: needs to be more general here
  if alpha == []:
    # # one extra year in first axis to account for construction year (no production)
    alpha = np.zeros([n_projLife, n_hours_per_year])
    # it necessary to have alpha be [year, clusterhour] instead of [year, cluster, hour]
    #    clusterhour loops through hours first, then cluster
    signal = mdl.LMP[scenario_ind] if is_stochastic else mdl.LMP

    realized_alpha = [[signal[y][d][h] \
                          for d in mdl.set_days
                            for h in mdl.set_time] # order here matches *indeces*
                              for y in fullYearsArray[1:]] #shape here is [year, hour]
    # # first column of year axis is 0 for project year 0
    realized_alpha = np.array(realized_alpha)
    alpha[1:,:] = realized_alpha

  pcount = -1
  for p, pyear in enumerate(fullYearsArray):
    if pyear == 0:
      continue

    if pyear > fullYearsArray[p-1]:
      pcount +=1

    for time in range(n_hours_per_year):
      ind = tuple(indeces[pcount,time])
      # looping through all DISPATCHES variables pertaining to this specific dispatch
      #   e.g., turbine costs due to work done by turbine + compressor, separate variables
      dispatch_driver = 0
      for ds, dStr in enumerate(exprs):
        dispatch_driver += operator.attrgetter(dStr)(scenario.period[ind])[0] * mults[ds]

      # getting weights for each day/cluster
      dy, yr = ind[1:]
      weight = weights_days[yr][dy]  # extracting weight for year + day

      # storing individual Pyomo dispatch
      dispatch_array[p, time] = dispatch_driver * weight

  return alpha, dispatch_array

###################################
# Methods to create TEAL components
###################################

def createCapex(alpha, driver, projLife):
  """
    Constructs the TEAL Capex Cashflow
    @ In, alpha, float, price
    @ In, driver, Pyomo Expression, quantity used in cashflow
    @ In, projLife, float, component life
    @ Out, cf, TEAL.src.CashFlows.Component, cashflow sale for each capital expenditures
  """
  # extract alpha, driver as just one value
  cf = CashFlows.Capex()
  cf.name = 'Cap'
  # life = comp._lifetime
  cf.initParams(projLife)
  cfParams = {'name': 'Cap',
               'alpha': alpha,
               'driver': driver,
               'reference': 1.0,
               'X': 1.0,
               'depreciate': 1,
               'mult_target': None,
               'inflation': False,
              }
  cf.setParams(cfParams)
  return cf

def createRecurringYearly(alpha, driver, lifeVector):
  """
    Constructs a TEAL Yearly Cashflow
    @ In, alpha, float, yearly price/cost to populate
    @ In, driver, Pyomo Expression, quantity used in cashflow
    @ In, lifeVector, numpy array, years in project life
    @ Out, cf, TEAL.src.CashFlows.Component, cashflow sale for the recurring yearly
  """
  lifeVector = np.hstack([0, lifeVector])
  cf = CashFlows.Recurring()
  cfParams = {'name': 'FixedOM',
               'X': 1,
               'mult_target': None,
               'inflation': False}
  cf.setParams(cfParams)

  # convert to binary mask, includes a zero for year 0 (construction)
  projYears = np.array([y>1 for y in lifeVector], dtype=int)
  projYears = projYears.astype(object)

  # 0 for first year (build year) -> TODO couldn't this be automatic?
  alphas  = projYears * alpha
  drivers = projYears * driver

  # construct annual summary cashflows
  cf.computeYearlyCashflow(alphas, drivers)
  return cf

def createRecurringHourly(alpha, driver, projLife):
  """
    Constructs a TEAL Hourly Cashflow
    @ In, alpha, float, price
    @ In, driver, Pyomo Expression, quantity used in cashflow
    @ In, projLife, float, component life
    @ Out, cf, TEAL.src.CashFlows.Component, cashflow sale for each capital expenditures
  """
  projLife += 1
  cf = CashFlows.Recurring()
  cfParams = {'name': 'Hourly',
               'X': 1,
               'mult_target': None,
               'inflation': False}
  cf.setParams(cfParams)
  cf.initParams(projLife, pyomoVar=True)
  for year in range(projLife):
    if isinstance(alpha, float):
      cf.computeIntrayearCashflow(year, alpha, driver[year, :])
    else:
      cf.computeIntrayearCashflow(year, alpha[year, :], driver[year, :])
  return cf



#################################################################################
# DISPATCHES was produced under the DOE Design Integration and Synthesis Platform
# to Advance Tightly Coupled Hybrid Energy Systems program (DISPATCHES), and is
# copyright (c) 2020-2023 by the software owners: The Regents of the University
# of California, through Lawrence Berkeley National Laboratory, National
# Technology & Engineering Solutions of Sandia, LLC, Alliance for Sustainable
# Energy, LLC, Battelle Energy Alliance, LLC, University of Notre Dame du Lac, et
# al. All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. Both files are also available online at the URL:
# "https://github.com/gmlc-dispatches/dispatches".
#################################################################################
"""
Ideal property package for H2 vapor
"""
# Import Python libraries
import logging

from pyomo.environ import units as pyunits

# Import IDAES cores
from idaes.core import VaporPhase, Component, PhaseType as PT

from idaes.models.properties.modular_properties.state_definitions import FTPx
from idaes.models.properties.modular_properties.eos.ideal import Ideal

from idaes.models.properties.modular_properties.pure.NIST import NIST

# Set up logger
_log = logging.getLogger(__name__)


# ---------------------------------------------------------------------
# Configuration dictionary for an ideal Benzene-Toluene system

# Data Sources:
# [1] The Properties of Gases and Liquids (1987)
#     4th edition, Chemical Engineering Series - Robert C. Reid
# [2] The NIST Webbook, https://webbook.nist.gov/, retrieved 15th Dec 2020

configuration = {
    # Specifying components
    "components": {
        'hydrogen': {"type": Component,
                     "valid_phase_types": PT.vaporPhase,
                     "elemental_composition": {"H": 2},
                     "cp_mol_ig_comp": NIST,
                     "enth_mol_ig_comp": NIST,
                     "entr_mol_ig_comp": NIST,
                     "parameter_data": {
                         "mw": (2.016e-3, pyunits.kg/pyunits.mol),  # [1]
                         "pressure_crit": (12.9e5, pyunits.Pa),  # [1]
                         "temperature_crit": (33.0, pyunits.K),  # [1]
                         "cp_mol_ig_comp_coeff": {
                             'A': (33.066178,  # [2]
                                   pyunits.J/pyunits.mol/pyunits.K),
                             'B': (-11.363417,  # [2]
                                   pyunits.J/pyunits.mol/pyunits.K/pyunits.kK),
                             'C': (11.432816,  # [2]
                                   pyunits.J/pyunits.mol/pyunits.K/pyunits.kiloK**2),
                             'D': (-2.772874,  # [2]
                                   pyunits.J/pyunits.mol/pyunits.K/pyunits.kiloK**3),
                             'E': (-0.158558,  # [2]
                                   pyunits.J/pyunits.mol/pyunits.K*pyunits.kiloK**2),
                             'F': (-9.980797,  # [2]
                                   pyunits.kJ/pyunits.mol),
                             'G': (172.707974,  # [2]
                                   pyunits.J/pyunits.mol/pyunits.K),
                             'H': (0,  # [2]
                                   pyunits.kJ/pyunits.mol)}}}},

    # Specifying phases
    "phases":  {'Vap': {"type": VaporPhase,
                        "equation_of_state": Ideal}},

    # Set base units of measurement
    "base_units": {"time": pyunits.s,
                   "length": pyunits.m,
                   "mass": pyunits.kg,
                   "amount": pyunits.mol,
                   "temperature": pyunits.K},

    # Specifying state definition
    "state_definition": FTPx,
    "state_bounds": {"flow_mol": (0, 100, 100000, pyunits.mol/pyunits.s),
                     "temperature": (273.15, 300, 1000, pyunits.K),
                     "pressure": (5e4, 1e5, 1e6, pyunits.Pa)},
    "pressure_ref": (101325, pyunits.Pa),  # [2]
    "temperature_ref": (298.15, pyunits.K)}  # [2]



#################################################################################
# DISPATCHES was produced under the DOE Design Integration and Synthesis Platform
# to Advance Tightly Coupled Hybrid Energy Systems program (DISPATCHES), and is
# copyright (c) 2020-2023 by the software owners: The Regents of the University
# of California, through Lawrence Berkeley National Laboratory, National
# Technology & Engineering Solutions of Sandia, LLC, Alliance for Sustainable
# Energy, LLC, Battelle Energy Alliance, LLC, University of Notre Dame du Lac, et
# al. All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. Both files are also available online at the URL:
# "https://github.com/gmlc-dispatches/dispatches".
#################################################################################

"""
This is an integrated model for the ultra-supercritical power plant with
the thermal energy storage system. The optimal design decisions obtained
from the solution of charge and discharge design models are used to
integrate the energy storage system with the power plant.
In this implementation, the optimal discrete choices are:
(1) Storage fluid: Solar salt
(2) Steam source during charge: HP steam
(3) Condensate recycle: Boiler Feed Pump
(4) Condensate source during discharge: Boiler Feed Pump

In addition, in this integrate model, both the charge and discharge
heat exchangers are included in the same flowsheet. The resulting model is
a nonlinear programming model and is solved using IPOPT.
"""

__author__ = "Naresh Susarla and Soraya Rawlings"

# Import Python libraries
from math import pi
import logging

# Import Pyomo libraries
from pyomo.environ import (Param, Constraint, Objective, Reals,
                           NonNegativeReals, TransformationFactory, Expression,
                           maximize, RangeSet, value, log, Var)
from pyomo.environ import units as pyunits
from pyomo.network import Arc
from pyomo.util.calc_var_value import calculate_variable_from_constraint

# Import IDAES Core libraries
from idaes.core.util import model_serializer as ms
from idaes.core import MaterialBalanceType
from idaes.core.util.initialization import propagate_state
from idaes.core.solvers import get_solver
from idaes.core.util.model_statistics import degrees_of_freedom
import idaes.core.util.scaling as iscale
import idaes.logger as idaeslog

# Import IDAES Unit Model Libraries
from idaes.models.unit_models import (HeatExchanger,
                                      MomentumMixingType,
                                      PressureChanger,
                                      Heater)
from idaes.models_extra.power_generation.unit_models.helm import (
    HelmMixer,
    HelmTurbineStage,
    HelmSplitter
)
from idaes.models.unit_models.heat_exchanger import (
    delta_temperature_underwood_callback)
from idaes.models.unit_models.pressure_changer import (
    ThermodynamicAssumption)

# Import DISPATCHES libraries
from dispatches.case_studies.fossil_case.ultra_supercritical_plant import (
    ultra_supercritical_powerplant as usc)
from dispatches.properties import solarsalt_properties
from idaes.core.util.exceptions import ConfigurationError
logging.basicConfig(level=logging.INFO)
logging.getLogger('pyomo.repn.plugins.nl_writer').setLevel(logging.ERROR)


def create_integrated_model(m, max_power=None):
    """This method uses the ultra-supercritical power plant model to
    integrate a thermal energy storage (TES) system. The unit models
    required for the TES are instantiated and connected to the power
    plant model using the Arc objects.

    """

    # Add Solar salt properties
    m.fs.solar_salt_properties = solarsalt_properties.SolarsaltParameterBlock()

    ###########################################################################
    #  Add storage splitters
    ###########################################################################
    # hp_spit to divert some steam from high pressure inlet during charge
    # ess_bfp_split to divert some condensate from bf pump during discharge
    m.fs.ess_hp_split = HelmSplitter(
        property_package=m.fs.prop_water,
        outlet_list=["to_hxc", "to_turbine"],
    )
    m.fs.ess_bfp_split = HelmSplitter(
        property_package=m.fs.prop_water,
        outlet_list=["to_hxd", "to_recyclemix"],
    )

    ###########################################################################
    #  Add cooler and hx pump
    ###########################################################################
    # To ensure the outlet of charge heat exchanger is a subcooled
    # liquid before mixing it with the plant, a cooler is added after
    # the heat exchanger
    m.fs.cooler = Heater(
        property_package=m.fs.prop_water,
        has_pressure_change=True,
    )

    # A pump, if needed, is used to increase the pressure of the water
    # to allow mixing it at a desired location within the plant
    m.fs.hx_pump = PressureChanger(
        property_package=m.fs.prop_water,
        material_balance_type=MaterialBalanceType.componentTotal,
        thermodynamic_assumption=ThermodynamicAssumption.pump,
    )

    ###########################################################################
    #  Add recycle mixer
    ###########################################################################
    m.fs.recycle_mixer = HelmMixer(
        momentum_mixing_type=MomentumMixingType.none,
        inlet_list=["from_bfw_out", "from_hx_pump"],
        property_package=m.fs.prop_water,
    )

    # Add charge heat exchanger
    m.fs.hxc = HeatExchanger(
        delta_temperature_callback=delta_temperature_underwood_callback,
        hot_side_name="shell",
        cold_side_name="tube",
        shell={"property_package": m.fs.prop_water},
        tube={"property_package": m.fs.solar_salt_properties},
    )

    # Add discharge heat exchanger
    m.fs.hxd = HeatExchanger(
        delta_temperature_callback=delta_temperature_underwood_callback,
        hot_side_name="shell",
        cold_side_name="tube",
        shell={"property_package": m.fs.solar_salt_properties},
        tube={"property_package": m.fs.prop_water},
    )

    ###########################################################################
    #  Design of Storage Heat Exchanger
    ###########################################################################

    # -------Shell-n-tube counter-flow heat exchanger design parameters-------
    m.fs.data_storage_hx = {
        'tube_thickness': 0.004,
        'tube_inner_dia': 0.032,
        'tube_outer_dia': 0.036,
        'k_steel': 21.5,
        'number_tubes': 20,
        'shell_inner_dia': 1
    }

    m.fs.tube_thickness = Param(
        initialize=m.fs.data_storage_hx['tube_thickness'],
        doc='Tube thickness in m')
    m.fs.tube_inner_dia = Param(
        initialize=m.fs.data_storage_hx['tube_inner_dia'],
        doc='Tube inner diameter in m')
    m.fs.tube_outer_dia = Param(
        initialize=m.fs.data_storage_hx['tube_outer_dia'],
        doc='Tube outer diameter in m')
    m.fs.k_steel = Param(
        initialize=m.fs.data_storage_hx['k_steel'],
        doc='Thermal conductivity of steel in W/m.K')
    m.fs.n_tubes = Param(
        initialize=m.fs.data_storage_hx['number_tubes'],
        doc='Number of tubes')
    m.fs.shell_inner_dia = Param(
        initialize=m.fs.data_storage_hx['shell_inner_dia'],
        doc='Shell inner diameter in m')
    m.fs.tube_cs_area = Expression(
        expr=(pi / 4) *
        (m.fs.tube_inner_dia ** 2),
        doc="Tube cross sectional area")
    m.fs.tube_out_area = Expression(
        expr=(pi / 4) *
        (m.fs.tube_outer_dia ** 2),
        doc="Tube cross sectional area including thickness in m2")
    m.fs.shell_eff_area = Expression(
        expr=(
            (pi / 4) *
            (m.fs.shell_inner_dia ** 2) -
            m.fs.n_tubes *
            m.fs.tube_out_area),
        doc="Effective shell cross sectional area in m2")

    m.fs.tube_dia_ratio = (m.fs.tube_outer_dia / m.fs.tube_inner_dia)
    m.fs.log_tube_dia_ratio = log(m.fs.tube_dia_ratio)

    # Data to compute overall heat transfer coefficient for the charge
    # heat exchanger using the Sieder-Tate Correlation. Parameters for
    # tube diameter and thickness assumed from the data in (2017) He
    # et al., Energy Procedia 105, 980-985

    # -------- Charge Heat Exchanger Heat Transfer Coefficient --------
    m.fs.hxc.salt_reynolds_number = Expression(
        expr=(
            (m.fs.hxc.tube_inlet.flow_mass[0] *
             m.fs.tube_outer_dia) /
            (m.fs.shell_eff_area *
             m.fs.hxc.cold_side.properties_in[0].visc_d_phase["Liq"])
        ),
        doc="Salt Reynolds Number")
    m.fs.hxc.salt_prandtl_number = Expression(
        expr=(
            m.fs.hxc.cold_side.properties_in[0].cp_mass["Liq"] *
            m.fs.hxc.cold_side.properties_in[0].visc_d_phase["Liq"] /
            m.fs.hxc.cold_side.properties_in[0].therm_cond_phase["Liq"]
        ),
        doc="Salt Prandtl Number")
    m.fs.hxc.salt_prandtl_wall = Expression(
        expr=(
            m.fs.hxc.cold_side.properties_out[0].cp_mass["Liq"] *
            m.fs.hxc.cold_side.properties_out[0].visc_d_phase["Liq"] /
            m.fs.hxc.cold_side.properties_out[0].therm_cond_phase["Liq"]
        ),
        doc="Salt Prandtl Number at wall")
    m.fs.hxc.salt_nusselt_number = Expression(
        expr=(
            0.35 *
            (m.fs.hxc.salt_reynolds_number**0.6) *
            (m.fs.hxc.salt_prandtl_number**0.4) *
            ((m.fs.hxc.salt_prandtl_number /
              m.fs.hxc.salt_prandtl_wall) ** 0.25) *
            (2**0.2)
        ),
        doc="Salt Nusslet Number from 2019, App Ener (233-234), 126")
    m.fs.hxc.steam_reynolds_number = Expression(
        expr=(
            m.fs.hxc.shell_inlet.flow_mol[0] *
            m.fs.hxc.hot_side.properties_in[0].mw *
            m.fs.tube_inner_dia /
            (m.fs.tube_cs_area *
             m.fs.n_tubes *
             m.fs.hxc.hot_side.properties_in[0].visc_d_phase["Vap"])
        ),
        doc="Steam Reynolds Number")
    m.fs.hxc.steam_prandtl_number = Expression(
        expr=(
            (m.fs.hxc.hot_side.properties_in[0].cp_mol /
             m.fs.hxc.hot_side.properties_in[0].mw) *
            m.fs.hxc.hot_side.properties_in[0].visc_d_phase["Vap"] /
            m.fs.hxc.hot_side.properties_in[0].therm_cond_phase["Vap"]
        ),
        doc="Steam Prandtl Number")
    m.fs.hxc.steam_nusselt_number = Expression(
        expr=(
            0.023 *
            (m.fs.hxc.steam_reynolds_number**0.8) *
            (m.fs.hxc.steam_prandtl_number**(0.33)) *
            ((m.fs.hxc.hot_side.properties_in[0].visc_d_phase["Vap"] /
              m.fs.hxc.hot_side.properties_out[0].visc_d_phase["Liq"]) ** 0.14)
        ),
        doc="Steam Nusslet Number from 2001 Zavoico, Sandia")

    # Calculate heat transfer coefficients for the salt and steam
    # sides of charge heat exchanger
    m.fs.hxc.h_salt = Expression(
        expr=(
            m.fs.hxc.cold_side.properties_in[0].therm_cond_phase["Liq"] *
            m.fs.hxc.salt_nusselt_number /
            m.fs.tube_outer_dia
        ),
        doc="Salt side convective heat transfer coefficient in W/m.K")
    m.fs.hxc.h_steam = Expression(
        expr=(
            m.fs.hxc.hot_side.properties_in[0].therm_cond_phase["Vap"] *
            m.fs.hxc.steam_nusselt_number /
            m.fs.tube_inner_dia
        ),
        doc="Steam side convective heat transfer coefficient in W/m.K")

    @m.fs.hxc.Constraint(m.fs.time)
    def constraint_hxc_ohtc(b, t):
        return (
            m.fs.hxc.overall_heat_transfer_coefficient[t] *
            (2 * m.fs.k_steel *
             m.fs.hxc.h_steam +
             m.fs.tube_outer_dia *
             m.fs.log_tube_dia_ratio *
             m.fs.hxc.h_salt *
             m.fs.hxc.h_steam +
             m.fs.tube_dia_ratio *
             m.fs.hxc.h_salt *
             2 * m.fs.k_steel)
        ) == (2 * m.fs.k_steel *
              m.fs.hxc.h_salt *
              m.fs.hxc.h_steam)

    m.fs.hxc_to_cooler = Arc(
        source=m.fs.hxc.shell_outlet,
        destination=m.fs.cooler.inlet,
        doc="Connection from cooler to solar charge heat exchanger"
    )

    # ------- Discharge Heat Exchanger Heat Transfer Coefficient -------
    # Discharge heat exchanger salt and steam side constraints to
    # calculate Reynolds number, Prandtl number, and Nusselt number
    m.fs.hxd.salt_reynolds_number = Expression(
        expr=(
            m.fs.hxd.shell_inlet.flow_mass[0]
            * m.fs.tube_outer_dia
            / (m.fs.shell_eff_area
               * m.fs.hxd.hot_side.properties_in[0].visc_d_phase["Liq"])
        ),
        doc="Salt Reynolds Number"
    )
    m.fs.hxd.salt_prandtl_number = Expression(
        expr=(
            m.fs.hxd.hot_side.properties_in[0].cp_mass["Liq"]
            * m.fs.hxd.hot_side.properties_in[0].visc_d_phase["Liq"]
            / m.fs.hxd.hot_side.properties_in[0].therm_cond_phase["Liq"]
        ),
        doc="Salt Prandtl Number"
    )
    # Assuming that the wall conditions are same as those at the outlet
    m.fs.hxd.salt_prandtl_wall = Expression(
        expr=(
            m.fs.hxd.hot_side.properties_out[0].cp_mass["Liq"]
            * m.fs.hxd.hot_side.properties_out[0].visc_d_phase["Liq"]
            / m.fs.hxd.hot_side.properties_out[0].therm_cond_phase["Liq"]
        ),
        doc="Wall Salt Prandtl Number"
    )
    m.fs.hxd.salt_nusselt_number = Expression(
        expr=(
            0.35 * (m.fs.hxd.salt_reynolds_number**0.6)
            * (m.fs.hxd.salt_prandtl_number**0.4)
            * ((m.fs.hxd.salt_prandtl_number
                / m.fs.hxd.salt_prandtl_wall)**0.25)
            * (2**0.2)
        ),
        doc="Solar Salt Nusslet Number from 2019, App Ener (233-234), 126"
    )
    m.fs.hxd.steam_reynolds_number = Expression(
        expr=(
            m.fs.hxd.tube_inlet.flow_mol[0]
            * m.fs.hxd.cold_side.properties_in[0].mw
            * m.fs.tube_inner_dia
            / (m.fs.tube_cs_area
               * m.fs.n_tubes
               * m.fs.hxd.cold_side.properties_in[0].visc_d_phase["Liq"])
        ),
        doc="Steam Reynolds Number"
    )
    m.fs.hxd.steam_prandtl_number = Expression(
        expr=(
            (m.fs.hxd.cold_side.properties_in[0].cp_mol
             / m.fs.hxd.cold_side.properties_in[0].mw)
            * m.fs.hxd.cold_side.properties_in[0].visc_d_phase["Liq"]
            / m.fs.hxd.cold_side.properties_in[0].therm_cond_phase["Liq"]
        ),
        doc="Steam Prandtl Number"
    )
    m.fs.hxd.steam_nusselt_number = Expression(
        expr=(
            0.023 * (m.fs.hxd.steam_reynolds_number ** 0.8)
            * (m.fs.hxd.steam_prandtl_number ** (0.33))
            * ((m.fs.hxd.cold_side.properties_in[0].visc_d_phase["Liq"]
                / m.fs.hxd.cold_side.properties_out[0].visc_d_phase["Vap"]
                ) ** 0.14)
        ),
        doc="Steam Nusslet Number from 2001 Zavoico, Sandia"
    )

    # Discharge heat exchanger salt and steam side heat transfer
    # coefficients
    m.fs.hxd.h_salt = Expression(
        expr=(
            m.fs.hxd.hot_side.properties_in[0].therm_cond_phase["Liq"]
            * m.fs.hxd.salt_nusselt_number / m.fs.tube_outer_dia
        ),
        doc="Salt side convective heat transfer coefficient in W/m.K"
    )
    m.fs.hxd.h_steam = Expression(
        expr=(
            m.fs.hxd.cold_side.properties_in[0].therm_cond_phase["Liq"]
            * m.fs.hxd.steam_nusselt_number / m.fs.tube_inner_dia
        ),
        doc="Steam side convective heat transfer coefficient in W/m.K"
    )

    @m.fs.hxd.Constraint(m.fs.time,
                         doc="Overall heat transfer coefficient for hxd")
    def constraint_hxd_ohtc(b, t):
        return (
            m.fs.hxd.overall_heat_transfer_coefficient[t] *
            (2 * m.fs.k_steel *
             m.fs.hxd.h_steam +
             m.fs.tube_outer_dia *
             m.fs.log_tube_dia_ratio *
             m.fs.hxd.h_salt *
             m.fs.hxd.h_steam +
             m.fs.tube_dia_ratio *
             m.fs.hxd.h_salt *
             2 * m.fs.k_steel)
        ) == (2 * m.fs.k_steel *
              m.fs.hxd.h_salt *
              m.fs.hxd.h_steam)

    m.fs.es_turbine = HelmTurbineStage(
        property_package=m.fs.prop_water,
    )
    ###########################################################################
    #  Add constraints to model
    ###########################################################################
    _make_constraints(m, max_power=max_power)

    ###########################################################################
    #  Create the stream Arcs and return the model
    ###########################################################################
    _create_arcs(m)
    TransformationFactory("network.expand_arcs").apply_to(m)

    return m


def _make_constraints(m, max_power=None):
    """Declare flowsheet constraints for the integrated model
    """

    # Cooler
    @m.fs.cooler.Constraint(m.fs.time,
                            doc="Cooler outlet temperature to be subcooled")
    def constraint_cooler_enth2(b, t):
        return (
            b.control_volume.properties_out[t].temperature <=
            (b.control_volume.properties_out[t].temperature_sat - 5 *pyunits.K)
        )

    # HX pump
    @m.fs.Constraint(m.fs.time,
                     doc="HX pump out pressure equal to BFP out pressure")
    def constraint_hxpump_presout(b, t):
        return (m.fs.hx_pump.outlet.pressure[t] ==
                m.main_steam_pressure * 1.1231 * pyunits.Pa)

    # Recycle mixer
    @m.fs.recycle_mixer.Constraint(
        m.fs.time,
        doc="Recycle mixer outlet pressure equal to min inlet pressure")
    def recyclemixer_pressure_constraint(b, t):
        return b.from_bfw_out_state[t].pressure == b.mixed_state[t].pressure

    m.fs.production_cons.deactivate()

    @m.fs.Constraint(m.fs.time)
    def production_cons_with_storage(b, t):
        return (
            (-1 * sum(m.fs.turbine[p].work_mechanical[t]
                      for p in m.set_turbine)
             - m.fs.hx_pump.control_volume.work[0]
             ) ==
            m.fs.plant_power_out[t] * 1e6 * (pyunits.W/pyunits.MW)
        )

    m.fs.net_power = Expression(
        expr=(m.fs.plant_power_out[0]
              + (-1e-6 * (pyunits.MW/pyunits.W))
              * m.fs.es_turbine.work_mechanical[0])
    )

    m.fs.max_boiler_duty = Param(
        initialize=940,
        mutable=False,
        units=pyunits.MW,
        doc='Maximum thermal power of the boiler at maximum electric plant power in MW')

    m.fs.boiler_eff = Expression(
        expr=0.2143 * (m.fs.plant_heat_duty[0] / m.fs.max_boiler_duty)
        + 0.7357,
        doc="Boiler efficiency in fraction"
    )

    m.fs.coal_heat_duty = Var(
        initialize=1000,
        bounds=(0, 1e5),
        units=pyunits.MW,
        doc="Coal thermal power supplied to boiler in MW")

    def coal_heat_duty_rule(b):
        return m.fs.coal_heat_duty * m.fs.boiler_eff == (
            m.fs.plant_heat_duty[0])
    m.fs.coal_heat_duty_eq = Constraint(rule=coal_heat_duty_rule)

    m.fs.cycle_efficiency = Expression(
        expr=m.fs.net_power / m.fs.coal_heat_duty * 100,
        doc="Cycle efficiency in %"
    )


def _create_arcs(m):
    """Create arcs to connect TES to the flowsheet
    """

    # Disconnect arcs from ultra supercritical plant base model to
    # connect the charge heat exchanger
    for arc_s in [m.fs.bfp_to_fwh8,
                  m.fs.rh1_to_turb3]:
        arc_s.expanded_block.enth_mol_equality.deactivate()
        arc_s.expanded_block.flow_mol_equality.deactivate()
        arc_s.expanded_block.pressure_equality.deactivate()

    m.fs.rh1_to_esshp = Arc(
        source=m.fs.reheater[1].outlet,
        destination=m.fs.ess_hp_split.inlet,
        doc="Connection from reheater to IP splitter"
    )
    m.fs.esshp_to_turb3 = Arc(
        source=m.fs.ess_hp_split.to_turbine,
        destination=m.fs.turbine[3].inlet,
        doc="Connection from HP splitter to turbine 3"
    )
    m.fs.esshp_to_hxc = Arc(
        source=m.fs.ess_hp_split.to_hxc,
        destination=m.fs.hxc.shell_inlet,
        doc="Connection from HP splitter to HXC inlet 1"
    )
    m.fs.cooler_to_hxpump = Arc(
        source=m.fs.cooler.outlet,
        destination=m.fs.hx_pump.inlet,
        doc="Connection from cooler to HX pump"
    )
    m.fs.hxpump_to_recyclemix = Arc(
        source=m.fs.hx_pump.outlet,
        destination=m.fs.recycle_mixer.from_hx_pump,
        doc="Connection from HX pump to recycle mixer"
    )
    m.fs.bfp_to_essbfp = Arc(
        source=m.fs.bfp.outlet,
        destination=m.fs.ess_bfp_split.inlet,
        doc="Connection from BFP outlet to discharge split"
    )
    m.fs.essbfp_to_hxd = Arc(
        source=m.fs.ess_bfp_split.to_hxd,
        destination=m.fs.hxd.tube_inlet,
        doc="Connection from BFP outlet to discharge split"
    )
    m.fs.essbfp_to_recyclemix = Arc(
        source=m.fs.ess_bfp_split.to_recyclemix,
        destination=m.fs.recycle_mixer.from_bfw_out,
        doc="Connection from BFP outlet to recycle mixer"
    )
    m.fs.hxd_to_esturbine = Arc(
        source=m.fs.hxd.tube_outlet,
        destination=m.fs.es_turbine.inlet,
        doc="Connection from BFP outlet to discharge split"
    )
    m.fs.recyclemix_to_fwh8 = Arc(
        source=m.fs.recycle_mixer.outlet,
        destination=m.fs.fwh[8].tube_inlet,
        doc="Connection from Recycle Mixer to FWH8 tube side"
    )


def set_model_input(m):
    """Define model inputs and fixed variables or parameter values in the
    model to achieve a square problem, i.e. zero degrees of
    freedom.

    All the parameter values in this block, unless otherwise stated
    explicitly, are either assumed or estimated for a total power out
    of 437 MW

    """

    ###########################################################################
    #  Charge Heat Exchanger section
    ###########################################################################
    # Add heat exchanger area from supercritical plant
    # model_input. For conceptual design optimization, area is unfixed
    # and optimized
    m.fs.hxc.area.fix(2500)
    m.fs.hxd.area.fix(2000)

    # Define storage fluid conditions. The fluid inlet flow is fixed
    # during initialization, but is unfixed and determined during
    # optimization
    m.fs.hxc.tube_inlet.flow_mass.fix(140)
    m.fs.hxc.tube_inlet.temperature.fix(513.15)
    m.fs.hxc.tube_inlet.pressure.fix(101325)

    m.fs.hxd.shell_inlet.flow_mass.fix(200)
    m.fs.hxd.shell_inlet.temperature.fix(853.15)
    m.fs.hxd.shell_inlet.pressure.fix(101325)

    # Cooler outlet enthalpy is fixed during model build to ensure the
    # inlet to the pump is liquid. However, this is unfixed during
    # design optimization. The temperature is at the outlet of cooler
    # is constrained in the model
    m.fs.cooler.outlet.enth_mol[0].fix(10000)
    m.fs.cooler.deltaP[0].fix(0)

    # HX pump efficiecncy assumption
    m.fs.hx_pump.efficiency_pump.fix(0.80)

    m.fs.es_turbine.ratioP.fix(0.0286)
    m.fs.es_turbine.efficiency_isentropic.fix(0.8)
    ###########################################################################
    #  ESS VHP and HP splitters
    ###########################################################################
    # The model is built for a fixed flow of steam through the
    # charger.  This flow of steam to the charger is unfixed and
    # determine during design optimization
    m.fs.ess_hp_split.split_fraction[0, "to_hxc"].fix(0.1)
    m.fs.ess_bfp_split.split_fraction[0, "to_hxd"].fix(0.12)

    assert degrees_of_freedom(m) == 0


def set_scaling_factors(m):
    """Scaling factors in the flowsheet

    """

    # Include scaling factors
    for fluid in [m.fs.hxc, m.fs.hxd]:
        iscale.set_scaling_factor(fluid.area, 1e-2)
        iscale.set_scaling_factor(
            fluid.overall_heat_transfer_coefficient, 1e-3)
        iscale.set_scaling_factor(fluid.shell.heat, 1e-6)
        iscale.set_scaling_factor(fluid.tube.heat, 1e-6)

    iscale.set_scaling_factor(m.fs.hx_pump.control_volume.work, 1e-6)
    iscale.set_scaling_factor(m.fs.es_turbine.control_volume.work, 1e-6)

    for k in [m.fs.cooler]:
        iscale.set_scaling_factor(k.control_volume.heat, 1e-6)


def initialize(m, solver=None,
               outlvl=idaeslog.NOTSET,
               optarg={"tol": 1e-8, "max_iter": 300}):
    """Initialize the units included in the flowsheet
    """

    if solver is None:
        optarg = {
            "max_iter": 300,
            "halt_on_ampl_error": "yes",
        }
        solver = get_solver(solver, optarg)

    # Include scaling factors
    iscale.calculate_scaling_factors(m)

    # Initialize splitters
    propagate_state(m.fs.rh1_to_esshp)
    m.fs.ess_hp_split.initialize(outlvl=outlvl,
                                 optarg=solver.options)

    # Re-initialize turbines connected to splitters since the flow is
    # not the same as before
    propagate_state(m.fs.esshp_to_turb3)
    m.fs.turbine[3].initialize(outlvl=outlvl,
                               optarg=solver.options)

    # Initialize charge heat exchanger
    propagate_state(m.fs.esshp_to_hxc)
    m.fs.hxc.initialize(outlvl=outlvl,
                        optarg=solver.options)

    # Initialize cooler
    propagate_state(m.fs.hxc_to_cooler)
    m.fs.cooler.initialize(outlvl=outlvl,
                           optarg=solver.options)

    # Initialize HX pump
    propagate_state(m.fs.cooler_to_hxpump)
    m.fs.hx_pump.initialize(outlvl=outlvl,
                            optarg=solver.options)

    # Initialize BFP split
    propagate_state(m.fs.bfp_to_essbfp)
    m.fs.ess_bfp_split.initialize(outlvl=outlvl,
                                  optarg=solver.options)

    # Initialize recycle mixer
    propagate_state(m.fs.essbfp_to_recyclemix)
    propagate_state(m.fs.hxpump_to_recyclemix)
    m.fs.recycle_mixer.initialize(outlvl=outlvl)

    # Initialize recycle mixer
    propagate_state(m.fs.essbfp_to_hxd)
    m.fs.hxd.initialize(outlvl=outlvl,
                        optarg=solver.options)

    # Initialize discharge heat exchanger
    propagate_state(m.fs.hxd_to_esturbine)
    m.fs.es_turbine.initialize(outlvl=outlvl,
                               optarg=solver.options)

    # Check and raise an error if the degrees of freedom are not 0
    if not degrees_of_freedom(m) == 0:
        raise ConfigurationError(
            "The degrees of freedom after building the model are not 0. "
            "You have {} degrees of freedom. "
            "Please check your inputs to ensure a square problem "
            "before initializing the model.".format(degrees_of_freedom(m))
            )

    res = solver.solve(m, options=optarg)

    print("Integrated Model Initialization = ",
          res.solver.termination_condition)
    print("***************   Integrated Model Initialized   ***************")


def build_costing(m):
    """This method adds cost correlations for the storage design analysis
    and it is used to estimate the capital and operatig cost of
    integrating an energy storage system. It contains cost
    correlations to estimate the capital cost of charge and discharge
    heat exchangers, salt storage tank, molten salt pump, and salt
    inventory.

    """

    # All the computed capital costs are annualized. The operating
    # cost is for 1 year. In addition, operating savings in terms of
    # annual coal cost are estimated based on the differential
    # reduction of coal consumption as compared to ramped baseline
    # power plant. The cost correlations used here are taken from the
    # IDAES costing method.

    ###########################################################################
    #  Data                                                                   #
    ###########################################################################
    # Chemical engineering cost index for 2019
    m.CE_index = 607.5

    # The q baseline_charge corresponds to heat duty of a plant with
    # no storage and producing 400 MW power
    m.data_cost = {
        'coal_price': 2.11e-9,
        'cooling_price': 3.3e-9,
        'q_baseline_charge': 838565942.4732262,
        'solar_salt_price': 0.49,
        'hitec_salt_price': 0.93,
        'thermal_oil_price': 6.72,  # $/kg
        'storage_tank_material': 3.5,
        'storage_tank_insulation': 235,
        'storage_tank_foundation': 1210
    }
    m.data_salt_pump = {
        'FT': 1.5,
        'FM': 2.0,
        'head': 3.281*5,
        'motor_FT': 1,
        'nm': 1
    }
    m.data_storage_tank = {
        'LbyD': 0.325,
        'tank_thickness': 0.039,
        'material_density': 7800
    }

    # Main flowsheet operation data
    m.fs.coal_price = Param(
        initialize=m.data_cost['coal_price'],
        doc='Coal price based on HHV for Illinois No.6 (NETL Report) in $/J')
    m.fs.cooling_price = Param(
        initialize=m.data_cost['cooling_price'],
        doc='Cost of chilled water for cooler from Sieder et al. in $/J')
    m.fs.q_baseline = Param(
        initialize=m.data_cost['q_baseline_charge'],
        doc='Boiler duty in Wth @ 699MW for baseline plant with no storage')
    m.fs.solar_salt_price = Param(
        initialize=m.data_cost['solar_salt_price'],
        doc='Solar salt price in $/kg')

    ###########################################################################
    #  Operating hours                                                        #
    ###########################################################################
    m.fs.hours_per_day = Var(
        initialize=24,
        bounds=(0, 24),
        doc='Estimated number of hours of charging per day'
    )
    m.fs.hours_per_day.fix()

    # Define number of years over which the capital cost is annualized
    m.fs.num_of_years = Param(
        initialize=30,
        doc='Number of years for capital cost annualization')

    ###########################################################################
    #  Capital cost                                                           #
    ###########################################################################

    m.fs.salt_amount = Var(
        initialize=6739292,
        doc="Solar salt purchase cost in $"
    )
    m.fs.salt_amount.fix()

    m.fs.hxc_salt_design_flow = Param(
        initialize=312,
        doc='Design flow of salt through hxc in kg/s')
    m.fs.hxc_salt_design_density = Param(
        initialize=1937.36,
        doc='Design density of salt through hxc')

    m.fs.hxd_salt_design_flow = Param(
        initialize=362.2,
        doc='Design flow of salt through hxd')
    m.fs.hxd_salt_design_density = Param(
        initialize=1721.12,
        doc='Design density of salt through hxd')

    m.fs.capital_cost = Param(
        initialize=0.407655e6,
        doc="Annualized capital cost for solar salt in $/yr")

    ###########################################################################
    #  Annual operating cost
    ###########################################################################
    m.fs.operating_hours = Expression(
        expr=365 * 3600 * m.fs.hours_per_day,
        doc="Number of operating hours per year")
    m.fs.operating_cost = Var(
        initialize=1000000,
        bounds=(0, 1e12),
        doc="Operating cost in $/year")

    def op_cost_rule(b):
        return m.fs.operating_cost == (
            m.fs.operating_hours * m.fs.coal_price *
            (m.fs.coal_heat_duty * 1e6)
            - (m.fs.cooling_price * m.fs.operating_hours *
               m.fs.cooler.heat_duty[0])
        )
    m.fs.op_cost_eq = Constraint(rule=op_cost_rule)

    ###########################################################################
    #  Annual capital and operating cost for full plant
    ###########################################################################

    # Add variables and functions to calculate the plant capital cost
    # and variable and fixed operating costs
    m.fs.plant_capital_cost = Var(
        initialize=1000000,
        bounds=(0, 1e12),
        doc="Annualized capital cost for the plant in $/year")
    m.fs.plant_fixed_operating_cost = Var(
        initialize=1000000,
        bounds=(0, 1e12),
        doc="Plant fixed operating cost in $/year")
    m.fs.plant_variable_operating_cost = Var(
        initialize=1000000,
        bounds=(0, 1e12),
        doc="Plant variable operating cost in $/year")

    def plant_cap_cost_rule(b):
        return m.fs.plant_capital_cost == (
            (2688973 * m.fs.plant_power_out[0]
             + 618968072) /
            m.fs.num_of_years
        ) * (m.CE_index / 575.4)
    m.fs.plant_cap_cost_eq = Constraint(rule=plant_cap_cost_rule)

    def op_fixed_plant_cost_rule(b):
        return m.fs.plant_fixed_operating_cost == (
            (16657.5 * m.fs.plant_power_out[0]
             + 6109833.3) /
            m.fs.num_of_years
        ) * (m.CE_index / 575.4)
    m.fs.op_fixed_plant_cost_eq = Constraint(
        rule=op_fixed_plant_cost_rule)

    def op_variable_plant_cost_rule(b):
        return m.fs.plant_variable_operating_cost == (
            31754.7 * m.fs.plant_power_out[0]
        ) * (m.CE_index / 575.4)
    m.fs.op_variable_plant_cost_eq = Constraint(
        rule=op_variable_plant_cost_rule)

    return m


def initialize_with_costing(m, solver=None):

    # Create a solver object if it is not passed
    if solver is None:
        optarg = {
            "max_iter": 300,
            "halt_on_ampl_error": "yes",
        }
        solver = get_solver(solver, optarg)

    # Initialize operating cost
    calculate_variable_from_constraint(
        m.fs.operating_cost,
        m.fs.op_cost_eq)

    # Initialize capital cost of power plant
    calculate_variable_from_constraint(
        m.fs.plant_capital_cost,
        m.fs.plant_cap_cost_eq)

    # Initialize plant fixed and variable operating costs
    calculate_variable_from_constraint(
        m.fs.plant_fixed_operating_cost,
        m.fs.op_fixed_plant_cost_eq)
    calculate_variable_from_constraint(
        m.fs.plant_variable_operating_cost,
        m.fs.op_variable_plant_cost_eq)

    # Check and raise an error if the degrees of freedom are not 0
    if not degrees_of_freedom(m) == 0:
        raise ConfigurationError(
            "The degrees of freedom after building costing block are not 0. "
            "You have {} degrees of freedom. "
            "Please check your inputs to ensure a square problem "
            "before initializing the model.".format(degrees_of_freedom(m))
            )

    res = solver.solve(m, options=optarg)
    print("Cost Initialization = ",
          res.solver.termination_condition)
    print("******************** Costing Initialized *************************")
    print()
    print()


def add_bounds(m):
    """Add bounds to all units in thermal energy storage system

    Unless stated otherwise, the temperature is in K, pressure in
    Pa, flow in mol/s, massic flow in kg/s, and heat and heat duty
    in W

    """

    m.flow_max = m.main_flow * 3        # Units in mol/s
    m.flow_min = 11804                  # Units in mol/s
    m.salt_flow_max = 500               # Units in kg/s
    m.fs.heat_duty_max = 200e6          # Units in W

    # Add bounds to Solar salt charge heat exchanger
    m.fs.hxc.shell_inlet.flow_mol.setlb(0)
    m.fs.hxc.shell_inlet.flow_mol.setub(0.2 * m.flow_max)
    m.fs.hxc.tube_inlet.flow_mass.setlb(0)
    m.fs.hxc.tube_inlet.flow_mass.setub(m.salt_flow_max)
    m.fs.hxc.shell_outlet.flow_mol.setlb(0)
    m.fs.hxc.shell_outlet.flow_mol.setub(0.2 * m.flow_max)
    m.fs.hxc.tube_outlet.flow_mass.setlb(0)
    m.fs.hxc.tube_outlet.flow_mass.setub(m.salt_flow_max)
    m.fs.hxc.tube_inlet.pressure.setlb(101320)
    m.fs.hxc.tube_inlet.pressure.setub(101330)
    m.fs.hxc.tube_outlet.pressure.setlb(101320)
    m.fs.hxc.tube_outlet.pressure.setub(101330)
    m.fs.hxc.heat_duty.setlb(0)
    m.fs.hxc.heat_duty.setub(m.fs.heat_duty_max)
    m.fs.hxc.shell.heat.setlb(-m.fs.heat_duty_max)
    m.fs.hxc.shell.heat.setub(0)
    m.fs.hxc.tube.heat.setlb(0)
    m.fs.hxc.tube.heat.setub(m.fs.heat_duty_max)
    m.fs.hxc.tube.properties_in[0].enth_mass.setlb(0)
    m.fs.hxc.tube.properties_in[0].enth_mass.setub(1.5e6)
    m.fs.hxc.tube.properties_out[0].enth_mass.setlb(0)
    m.fs.hxc.tube.properties_out[0].enth_mass.setub(1.5e6)
    m.fs.hxc.overall_heat_transfer_coefficient.setlb(0)
    m.fs.hxc.overall_heat_transfer_coefficient.setub(10000)
    m.fs.hxc.area.setlb(0)
    m.fs.hxc.area.setub(6000)
    m.fs.hxc.delta_temperature_in.setlb(9)
    m.fs.hxc.delta_temperature_out.setlb(5)
    m.fs.hxc.delta_temperature_in.setub(80.5)
    m.fs.hxc.delta_temperature_out.setub(81)

    # Add bounds to Solar salt discharge heat exchanger
    m.fs.hxd.tube_inlet.flow_mol.setlb(0)
    m.fs.hxd.tube_inlet.flow_mol.setub(0.2 * m.flow_max)
    m.fs.hxd.shell_inlet.flow_mass.setlb(0)
    m.fs.hxd.shell_inlet.flow_mass.setub(m.salt_flow_max)
    m.fs.hxd.tube_outlet.flow_mol.setlb(0)
    m.fs.hxd.tube_outlet.flow_mol.setub(0.2 * m.flow_max)
    m.fs.hxd.shell_outlet.flow_mass.setlb(0)
    m.fs.hxd.shell_outlet.flow_mass.setub(m.salt_flow_max)
    m.fs.hxd.shell_inlet.pressure.setlb(101320)
    m.fs.hxd.shell_inlet.pressure.setub(101330)
    m.fs.hxd.shell_outlet.pressure.setlb(101320)
    m.fs.hxd.shell_outlet.pressure.setub(101330)
    m.fs.hxd.heat_duty.setlb(0)
    m.fs.hxd.heat_duty.setub(m.fs.heat_duty_max)
    m.fs.hxd.tube.heat.setub(m.fs.heat_duty_max)
    m.fs.hxd.tube.heat.setlb(0)
    m.fs.hxd.shell.heat.setub(0)
    m.fs.hxd.shell.heat.setlb(-m.fs.heat_duty_max)
    m.fs.hxd.shell.properties_in[0].enth_mass.setlb(0)
    m.fs.hxd.shell.properties_in[0].enth_mass.setub(1.5e6)
    m.fs.hxd.shell.properties_out[0].enth_mass.setlb(0)
    m.fs.hxd.shell.properties_out[0].enth_mass.setub(1.5e6)
    m.fs.hxd.overall_heat_transfer_coefficient.setlb(0)
    m.fs.hxd.overall_heat_transfer_coefficient.setub(10000)
    m.fs.hxd.area.setlb(0)
    m.fs.hxd.area.setub(6000)
    m.fs.hxd.delta_temperature_in.setlb(4.9)
    m.fs.hxd.delta_temperature_out.setlb(10)
    m.fs.hxd.delta_temperature_in.setub(300)
    m.fs.hxd.delta_temperature_out.setub(300)

    # Add bounds to the HX pump and Cooler
    for unit_k in [m.fs.hx_pump,
                   m.fs.cooler]:
        unit_k.inlet.flow_mol.setlb(0)
        unit_k.inlet.flow_mol.setub(0.2*m.flow_max)
        unit_k.outlet.flow_mol.setlb(0)
        unit_k.outlet.flow_mol.setub(0.2*m.flow_max)
    m.fs.cooler.heat_duty.setub(0)

    # Add bounds needed HP splitter
    for split in [m.fs.ess_hp_split]:
        split.to_hxc.flow_mol[:].setlb(0)
        split.to_hxc.flow_mol[:].setub(0.2 * m.flow_max)
        split.split_fraction[0.0, "to_hxc"].setlb(0)
        split.split_fraction[0.0, "to_hxc"].setub(1)
        split.split_fraction[0.0, "to_turbine"].setlb(0)
        split.split_fraction[0.0, "to_turbine"].setub(1)
        split.inlet.flow_mol[:].setlb(0)
        split.inlet.flow_mol[:].setub(m.flow_max)

    for split in [m.fs.ess_bfp_split]:
        split.to_hxd.flow_mol[:].setlb(0)
        split.to_hxd.flow_mol[:].setub(0.2 * m.flow_max)
        split.split_fraction[0.0, "to_hxd"].setlb(0)
        split.split_fraction[0.0, "to_hxd"].setub(1)
        split.split_fraction[0.0, "to_recyclemix"].setlb(0)
        split.split_fraction[0.0, "to_recyclemix"].setub(1)
        split.inlet.flow_mol[:].setlb(0)
        split.inlet.flow_mol[:].setub(m.flow_max)

    for mix in [m.fs.recycle_mixer]:
        mix.from_bfw_out.flow_mol.setlb(0)
        mix.from_bfw_out.flow_mol.setub(m.flow_max)
        mix.from_hx_pump.flow_mol.setlb(0)
        mix.from_hx_pump.flow_mol.setub(0.2 * m.flow_max)
        mix.outlet.flow_mol.setlb(0)
        mix.outlet.flow_mol.setub(m.flow_max)

    for k in m.set_turbine:
        m.fs.turbine[k].work.setlb(-1e10)
        m.fs.turbine[k].work.setub(0)
    m.fs.hx_pump.control_volume.work[0].setlb(0)
    m.fs.hx_pump.control_volume.work[0].setub(1e10)

    for unit_k in [m.fs.booster]:
        unit_k.inlet.flow_mol[:].setlb(0)
        unit_k.inlet.flow_mol[:].setub(m.flow_max)
        unit_k.outlet.flow_mol[:].setlb(0)
        unit_k.outlet.flow_mol[:].setub(m.flow_max)

    # Adding bounds on turbine splitters flow
    for k in m.set_turbine_splitter:
        m.fs.turbine_splitter[k].inlet.flow_mol[:].setlb(0)
        m.fs.turbine_splitter[k].inlet.flow_mol[:].setub(m.flow_max)
        m.fs.turbine_splitter[k].outlet_1.flow_mol[:].setlb(0)
        m.fs.turbine_splitter[k].outlet_1.flow_mol[:].setub(m.flow_max)
        m.fs.turbine_splitter[k].outlet_2.flow_mol[:].setlb(0)
        m.fs.turbine_splitter[k].outlet_2.flow_mol[:].setub(m.flow_max)

    return m


def main(max_power=None, load_from_file=None):

    if load_from_file is not None:

        # build plant model
        m = usc.build_plant_model()

        # Create a flowsheet, add properties, unit models, and arcs
        m = create_integrated_model(m, max_power=max_power)

        # Give all the required inputs to the model
        set_model_input(m)

        # Add scaling factor
        set_scaling_factors(m)

        # Add cost correlations
        m = build_costing(m)

        # Initialize with bounds
        ms.from_json(m, fname=load_from_file)
    else:

        m = usc.build_plant_model()
        usc.initialize(m)

        # Create a flowsheet, add properties, unit models, and arcs
        m = create_integrated_model(m, max_power=max_power)

        # Give all the required inputs to the model
        set_model_input(m)

        # Add scaling factor
        set_scaling_factors(m)

        # Initialize the model with a sequential initialization and custom
        # routines
        initialize(m)

        # Add cost correlations
        m = build_costing(m)

        # Initialize with bounds
        initialize_with_costing(m)

    # Add bounds
    add_bounds(m)

    return m


def print_results(m, results):

    print('================================')
    print('')
    print('')
    print("***************** Optimization Results ******************")
    print('Revenue ($/h): {:.6f}'.format(
        value(m.fs.revenue)))
    print('Storage Capital Cost ($/h): {:.6f}'.format(
        value(m.fs.capital_cost)/(365*24)))
    print('Fuel Cost ($/h): {:.6f}'.format(
        value(m.fs.operating_cost)/(365*24)))
    print('Hot Previous Salt Inventory (kg): {:.6f}'.format(
        value(m.fs.previous_salt_inventory_hot[0])))
    print('Hot Salt Inventory (kg): {:.6f}'.format(
        value(m.fs.salt_inventory_hot[0])))
    print('Cold Previous Salt Inventory (kg): {:.6f}'.format(
        value(m.fs.previous_salt_inventory_cold[0])))
    print('Cold Salt Inventory (kg): {:.6f}'.format(
        value(m.fs.salt_inventory_cold[0])))
    print('Salt Amount (kg): {:.6f}'.format(
        value(m.fs.salt_amount)))
    print('')
    print('')
    print("***************** Costing Results ******************")
    print('Obj (M$/year): {:.6f}'.format(value(m.obj)))
    print('Plant capital cost (M$/y): {:.6f}'.format(
        value(m.fs.plant_capital_cost) * 1e-6))
    print('Plant fixed operating costs (M$/y): {:.6f}'.format(
        value(m.fs.plant_fixed_operating_cost) * 1e-6))
    print('Plant variable operating costs (M$/y): {:.6f}'.format(
        value(m.fs.plant_variable_operating_cost) * 1e-6))
    print('Charge capital cost (M$/y): {:.6f}'.format(
        value(m.fs.capital_cost) * 1e-6))
    print('Charge Operating costs (M$/y): {:.6f}'.format(
        value(m.fs.operating_cost) * 1e-6))
    print('')
    print('')
    print("***************** Power Plant Operation ******************")
    print('')
    print('Net Power (MW): {:.6f}'.format(
        value(m.fs.net_power)))
    print('Plant Power (MW): {:.6f}'.format(
        value(m.fs.plant_power_out[0])))
    print('ES turbine Power (MW): {:.6f}'.format(
        value(m.fs.es_turbine.work[0]) * (-1e-6)))
    print('Boiler feed water flow (mol/s): {:.6f}'.format(
        value(m.fs.boiler.inlet.flow_mol[0])))
    print('Boiler duty (MW_th): {:.6f}'.format(
        value((m.fs.boiler.heat_duty[0]
               + m.fs.reheater[1].heat_duty[0]
               + m.fs.reheater[2].heat_duty[0])
              * 1e-6)))
    print('Cooling duty (MW_th): {:.6f}'.format(
        value(m.fs.cooler.heat_duty[0]) * -1e-6))
    print('HXC heat duty (MW): {:.6f}'.format(
        value(m.fs.hxc.heat_duty[0]) * 1e-6))
    print('HXD heat duty (MW): {:.6f}'.format(
        value(m.fs.hxd.heat_duty[0]) * 1e-6))
    print('Makeup water flow: {:.6f}'.format(
        value(m.fs.condenser_mix.makeup.flow_mol[0])))
    print('')
    print('')
    print("***************** Charge Heat Exchanger (HXC) ******************")
    print('')
    print('HXC area (m2): {:.6f}'.format(
        value(m.fs.hxc.area)))
    print('HXC Salt flow (kg/s): {:.6f}'.format(
        value(m.fs.hxc.tube_inlet.flow_mass[0])))
    print('HXC Salt temperature in (K): {:.6f}'.format(
        value(m.fs.hxc.tube_inlet.temperature[0])))
    print('HXC Salt temperature out (K): {:.6f}'.format(
        value(m.fs.hxc.tube_outlet.temperature[0])))
    print('HXC Steam flow to storage (mol/s): {:.6f}'.format(
        value(m.fs.hxc.shell_inlet.flow_mol[0])))
    print('HXC Water temperature in (K): {:.6f}'.format(
        value(m.fs.hxc.hot_side.properties_in[0].temperature)))
    print('HXC Steam temperature out (K): {:.6f}'.format(
        value(m.fs.hxc.hot_side.properties_out[0].temperature)))
    print('HXC Delta temperature at inlet (K): {:.6f}'.format(
        value(m.fs.hxc.delta_temperature_in[0])))
    print('HXC Delta temperature at outlet (K): {:.6f}'.format(
        value(m.fs.hxc.delta_temperature_out[0])))
    print('')
    print('')
    print("*************** Discharge Heat Exchanger (HXD) ****************")
    print('')
    print('HXD area (m2): {:.6f}'.format(
        value(m.fs.hxd.area)))
    print('HXD Salt flow (kg/s): {:.6f}'.format(
        value(m.fs.hxd.shell_inlet.flow_mass[0])))
    print('HXD Salt temperature in (K): {:.6f}'.format(
        value(m.fs.hxd.shell_inlet.temperature[0])))
    print('HXD Salt temperature out (K): {:.6f}'.format(
        value(m.fs.hxd.shell_outlet.temperature[0])))
    print('HXD Steam flow to storage (mol/s): {:.6f}'.format(
        value(m.fs.hxd.tube_inlet.flow_mol[0])))
    print('HXD Water temperature in (K): {:.6f}'.format(
        value(m.fs.hxd.cold_side.properties_in[0].temperature)))
    print('HXD Steam temperature out (K): {:.6f}'.format(
        value(m.fs.hxd.cold_side.properties_out[0].temperature)))
    print('HXD Delta temperature at inlet (K): {:.6f}'.format(
        value(m.fs.hxd.delta_temperature_in[0])))
    print('HXD Delta temperature at outlet (K): {:.6f}'.format(
        value(m.fs.hxd.delta_temperature_out[0])))
    print('')

    print('')
    print('Solver details')
    print(results)
    print(' ')
    print('==============================================================')


def print_reports(m):

    print('')
    for unit_k in [m.fs.boiler, m.fs.reheater[1],
                   m.fs.reheater[2],
                   m.fs.bfp, m.fs.bfpt,
                   m.fs.booster,
                   m.fs.condenser_mix,
                   m.fs.charge.hxc]:
        unit_k.display()

    for k in RangeSet(11):
        m.fs.turbine[k].report()
    for k in RangeSet(11):
        m.fs.turbine[k].display()
    for j in RangeSet(9):
        m.fs.fwh[j].report()
    for j in m.set_fwh_mixer:
        m.fs.fwh_mixer[j].display()


def model_analysis(m, solver,
                   power=None, max_power=None,
                   tank_scenario=None, fix_power=None):
    """Unfix variables for analysis. This section is deactived for the
    simulation of square model
    """

    # Add constraints and bounds for plant and discharge produced
    # power
    min_power = int(0.65 * max_power)# Units in MW
    max_power_storage = 29           # Units in MW
    min_power_storage = 1            # Units in MW

    if fix_power:
        m.fs.power_demand_eq = Constraint(
            expr=m.fs.net_power == power
        )
    else:
        m.fs.plant_power_min = Constraint(
            expr=m.fs.plant_power_out[0] >= min_power
        )
        m.fs.plant_power_max = Constraint(
            expr=m.fs.plant_power_out[0] <= max_power
        )
        m.fs.storage_power_min = Constraint(
            expr=m.fs.es_turbine.work[0] * (-1e-6) >= min_power_storage
        )
        m.fs.storage_power_max = Constraint(
            expr=m.fs.es_turbine.work[0] * (-1e-6) <= max_power_storage
        )

    # Add LMP signal value
    m.fs.lmp = Var(
        m.fs.time,
        domain=Reals,
        initialize=80,
        doc="Hourly LMP in $/MWh"
        )
    m.fs.lmp[0].fix(22)

    # Fix boiler outlet pressure and storage heat exchangers area and
    # salt temperatures
    m.fs.boiler.outlet.pressure.fix(m.main_steam_pressure)
    m.fs.salt_hot_temperature = 831
    m.fs.hxc.area.fix(1904)
    m.fs.hxd.area.fix(2830)
    m.fs.hxc.tube_outlet.temperature.fix(m.fs.salt_hot_temperature)
    m.fs.hxd.shell_inlet.temperature.fix(m.fs.salt_hot_temperature)
    m.fs.hxd.shell_outlet.temperature.fix(513.15)

    # Unfix variables that were fixed during initialization
    m.fs.boiler.inlet.flow_mol.unfix()
    m.fs.ess_hp_split.split_fraction[0, "to_hxc"].unfix()
    m.fs.ess_bfp_split.split_fraction[0, "to_hxd"].unfix()
    for salt_hxc in [m.fs.hxc]:
        salt_hxc.shell_inlet.unfix()
        salt_hxc.tube_inlet.flow_mass.unfix()
        salt_hxc.area.unfix()

    for salt_hxd in [m.fs.hxd]:
        salt_hxd.tube_inlet.unfix()
        salt_hxd.shell_inlet.flow_mass.unfix()
        salt_hxd.area.unfix()

    for unit in [m.fs.cooler]:
        unit.inlet.unfix()
    m.fs.cooler.outlet.enth_mol[0].unfix()

    # Add bounds to salt inventory
    inventory_max = 1e7          # Units in kg
    inventory_min = 75000        # Units in kg
    tank_max = 6739292           # Units in kg

    # Add salt inventory mass balances
    m.fs.previous_salt_inventory_hot = Var(
        m.fs.time,
        domain=NonNegativeReals,
        initialize=inventory_min,
        bounds=(0, inventory_max),
        doc="Hot salt at the beginning of time period in kg"
        )
    m.fs.salt_inventory_hot = Var(
        m.fs.time,
        domain=NonNegativeReals,
        initialize=inventory_min,
        bounds=(0, inventory_max),
        doc="Hot salt inventory at the end of the time period in kg"
        )
    m.fs.previous_salt_inventory_cold = Var(
        m.fs.time,
        domain=NonNegativeReals,
        initialize=tank_max-inventory_min,
        bounds=(0, inventory_max),
        doc="Cold salt at the beginning of the time period in kg"
        )
    m.fs.salt_inventory_cold = Var(
        m.fs.time,
        domain=NonNegativeReals,
        initialize=tank_max-inventory_min,
        bounds=(0, inventory_max),
        doc="Cold salt inventory at the end of the time period in kg"
        )

    @m.fs.Constraint(doc="Inventory balance at the end of the time period")
    def constraint_salt_inventory_hot(b):
        return (
            b.salt_inventory_hot[0] ==
            b.previous_salt_inventory_hot[0]
            + 3600 * b.hxc.tube_inlet.flow_mass[0]
            - 3600 * b.hxd.shell_inlet.flow_mass[0])

    @m.fs.Constraint(doc="Maximum salt inventory at any time")
    def constraint_salt_inventory(b):
        return (
            b.salt_inventory_hot[0] +
            b.salt_inventory_cold[0] == b.salt_amount)

    @m.fs.Constraint(doc="Maximum salt flow to hxd based on available hot salt")
    def constraint_salt_maxflow_hot(b):
        return (
            3600*m.fs.hxd.shell_inlet.flow_mass[0] <=
            m.fs.previous_salt_inventory_hot[0]
        )

    @m.fs.Constraint(doc="Maximum salt flow to hxc based on available cold salt")
    def constraint_salt_maxflow_cold(b):
        return (
            3600*m.fs.hxc.tube_inlet.flow_mass[0] <=
            m.fs.previous_salt_inventory_cold[0]
        )

    # Fix the previous salt inventory based on the tank scenario
    if tank_scenario == "hot_empty":
        m.fs.previous_salt_inventory_hot[0].fix(inventory_min)
        m.fs.previous_salt_inventory_cold[0].fix(tank_max - inventory_min)
    elif tank_scenario == "hot_half_full":
        m.fs.previous_salt_inventory_hot[0].fix(tank_max / 2)
        m.fs.previous_salt_inventory_cold[0].fix(tank_max / 2)
    elif tank_scenario == "hot_full":
        m.fs.previous_salt_inventory_hot[0].fix(tank_max - inventory_min)
        m.fs.previous_salt_inventory_cold[0].fix(inventory_min)
    else:
        print('Unknown scenario! Try hot_empty, hot_full, or hot_half_full')

    # Calculate revenue
    m.fs.revenue = Expression(
        expr=(m.fs.lmp[0] *
              m.fs.net_power),
        doc="Revenue function in $/hour assuming 1 hr operation"
    )

    # Add total cost as the objective function
    scaling_factor = 1e-2
    m.obj = Objective(
        expr=(
            m.fs.revenue
            - ((m.fs.operating_cost
                + m.fs.plant_fixed_operating_cost
                + m.fs.plant_variable_operating_cost) / (365 * 24))
        ) * scaling_factor,
        sense=maximize
    )


    # Solve the design optimization model
    results = solver.solve(
        m,
        tee=True,
        symbolic_solver_labels=True,
        options={
            "linear_solver": "ma27",
            "max_iter": 150
        }
    )

    print_results(m, results)
    
    return m


if __name__ == "__main__":

    optarg = {"max_iter": 300}
    solver = get_solver('ipopt', optarg)

    max_power = 436      # Units in MW
    power_demand = 460   # Units in MW

    # Tank scenarios: "hot_empty", "hot_full", "hot_half_full"
    tank_scenario = "hot_empty"

    # If hot_empty is selected and fix_power is True then ensure that
    # power_demand <= max_power
    fix_power = False

    # Build integrated ultra-supercritical plant model
    m_isp = main(max_power=max_power)

    # Solve the optimization problem
    m = model_analysis(m_isp,
                       solver,
                       power=power_demand,
                       max_power=max_power,
                       tank_scenario=tank_scenario,
                       fix_power=fix_power)



#################################################################################
# DISPATCHES was produced under the DOE Design Integration and Synthesis Platform
# to Advance Tightly Coupled Hybrid Energy Systems program (DISPATCHES), and is
# copyright (c) 2020-2023 by the software owners: The Regents of the University
# of California, through Lawrence Berkeley National Laboratory, National
# Technology & Engineering Solutions of Sandia, LLC, Alliance for Sustainable
# Energy, LLC, Battelle Energy Alliance, LLC, University of Notre Dame du Lac, et
# al. All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. Both files are also available online at the URL:
# "https://github.com/gmlc-dispatches/dispatches".
#################################################################################

__author__ = "Gabriel J. Soto, Anna Wrobel"

# This file contains utility functions for generating synthetic histories from a trained reduced
#    order model from RAVEN.
# Note that DISPATCHES, TEAL, and RAVEN are all assumed to be subdirectories within
#    the same directory.
from os import path
import ravenframework
from ravenframework.utils import xmlUtils
from ravenframework import ROMExternal
import numpy as np
import operator

###################
class SynHist_integration():
  """
    Class to facilitate integration of RAVEN's Synthetic History generation.
    Primary use is to load an already-trained price signal ARMA model. This model is
    built during the constructor of this class.
    Additional methods included for sampling from ARMA model and restructure data to accommodate
    DISPATCHES workflow.
  """

  def __init__(self, target_file):
    """
      Constructor for SynHist_integration.
      @ In, target_file, str, name of synthetic history file
      @ Out, None
    """
    self.target_file = target_file # full path for ARMA file
    self.inp = {'scaling': [1]} # this is a dummy input used for sampling call to ARMA model object
    self.runner = self.buildRunner() # builds object to run ARMA model sampling

  def buildRunner(self):
    """
      Builds synthetic history object from a pickledROM.
      Currently only ARMA models are supported for pickledROM.
    """
    if not path.exists(self.target_file):
      raise Exception(f"Target file not found at {self.target_file}")

    runner = ROMExternal.ROMLoader(self.target_file, ravenframework.__path__[0])

    nodes = []
    node = xmlUtils.newNode('ROM', attrib={'name': 'SyntheticHistory', 'subType': 'pickledRom'})
    node.append(xmlUtils.newNode('clusterEvalMode', text='clustered'))
    nodes.append(node)
    runner.setAdditionalParams(nodes)
    return runner

  def generateSyntheticHistory(self, signal_name, set_years):
    """
      Generates a sampled synthetic history from saved ROM object.
      @In, signal_name, str, name of signal in ARMA model
      @In, set_years, list, project years to simulate
      @Out, synHist, dict, dictionary including synthetic history and metadata
    """
    # sampling from ARMA model object
    synHist = self.runner.evaluate(self.inp)[0]
    newSynHist = {} # empty dictionary with restructured data

    # check that signal name included in synthetic history dictionary
    if signal_name not in synHist.keys():
      raise Exception(f"Signal name {signal_name} not found in sampled history keys: {synHist.keys()}")

    # extract actual synthetic history data
    synHistData = synHist[signal_name]

    # extract hourly, daily/cluster/ and year data arrays
    indexMap = synHist['_indexMap'][0][signal_name]
    if ('Time' in indexMap) or ('hour' in indexMap):
      hourKey = 'Time' if 'Time' in indexMap else 'hour'
    if '_ROM_Cluster' in indexMap:
      clusterKey = '_ROM_Cluster'
    if 'Year' in indexMap:
      yearKey = 'Year'

    # time sets from generated synthetic history
    synHistHours = np.asarray(synHist[hourKey]+1, dtype=int)
    synHistDays  = np.asarray(synHist[clusterKey]+1, dtype=int) # 20 clusters
    synHistYears = synHist[yearKey] # 2018-2045

    # checking that simulation years are included within synthetic history year set
    if not set(set_years).issubset(synHistYears):
      raise Exception(f"Years requested ({set_years}) not provided by ARMA Model: {synHistYears}")

    # getting cluster weights from ROM (deep hierarchy)
    #   using attrgetter to extract protected members -> self.runner.rom._segmentROM._macroSteps
    cluster_steps = operator.attrgetter("runner.rom._segmentROM._macroSteps")(self)
    newSynHist['weights_days'] = {}
    newSynHist['LMP'] = {}
    newSynHist['cluster_map'] = {}

    # loop through ROM years to extract clusters per year
    for year in set_years:
      newSynHist['weights_days'][year] = {}
      newSynHist['cluster_map'][year] = {}

      for cluster in synHistDays:
        # using attrgetter to extract protected members -> cluster_steps[year]._clusterInfo['map']
        cluster_map = operator.attrgetter('_clusterInfo')(cluster_steps[year])['map']
        cluster_ind = int(cluster-1)
        newSynHist['weights_days'][year][cluster] = len(cluster_map[cluster_ind])
        newSynHist['cluster_map'][year][cluster] = list(cluster_map[cluster_ind])

    # generated synthetic histories span multiple years, must index for given set_years
    year_index = [np.where(synHistYears == year)[0][0] for year in set_years]
    newSynHist['LMP'] = {year: {day: {hour: synHistData[year_ind, int(day-1), int(hour-1)]
                            for hour in synHistHours}
                  for day in synHistDays}
            for year_ind, year in zip(year_index, set_years)}

    return newSynHist



#################################################################################
# DISPATCHES was produced under the DOE Design Integration and Synthesis Platform
# to Advance Tightly Coupled Hybrid Energy Systems program (DISPATCHES), and is
# copyright (c) 2020-2023 by the software owners: The Regents of the University
# of California, through Lawrence Berkeley National Laboratory, National
# Technology & Engineering Solutions of Sandia, LLC, Alliance for Sustainable
# Energy, LLC, Battelle Energy Alliance, LLC, University of Notre Dame du Lac, et
# al. All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. Both files are also available online at the URL:
# "https://github.com/gmlc-dispatches/dispatches".
#################################################################################
"""
Test that headers are on all files
"""
# stdlib
from pathlib import Path
import os

import pytest

pytest.importorskip("addheader", reason="addheader (optional dev. dependency) not available")

# third-party
from addheader.add import FileFinder, detect_files
import yaml


@pytest.fixture
def package_root():
    """Determine package root.
    """
    import dispatches
    return Path(dispatches.__file__).parent


@pytest.fixture
def patterns(package_root):
    """Grab glob patterns from config file.
    """
    conf_file = package_root.parent / ".addheader.yml"
    if not conf_file.exists():
        print(f"Cannot load configuration file from '{conf_file}'. Perhaps this is not development mode?")
        return None
    with open(conf_file) as f:
        conf_data = yaml.safe_load(f)
    print(f"Patterns for finding files with headers: {conf_data['patterns']}")
    return conf_data["patterns"]


@pytest.mark.unit
def test_headers(package_root, patterns):
    if patterns is None:
        print(f"ERROR: Did not get glob patterns: skipping test")
    else:
        # modify patterns to match the files that should have headers
        ff = FileFinder(package_root, glob_patterns=patterns)
        has_header, missing_header = detect_files(ff)
        # ignore empty files (probably should add option in 'detect_files' for this)
        nonempty_missing_header = list(filter(lambda p: p.stat().st_size > 0, missing_header))
        #
        if len(nonempty_missing_header) > 0:
            pfx = str(package_root.resolve())
            pfx_len = len(pfx)
            file_list = ", ".join([str(p)[pfx_len + 1:] for p in nonempty_missing_header])
            print(f"Missing headers from files under '{pfx}{os.path.sep}': {file_list}")
        # uncomment to require all files to have headers
        assert len(nonempty_missing_header) == 0



#################################################################################
# DISPATCHES was produced under the DOE Design Integration and Synthesis Platform
# to Advance Tightly Coupled Hybrid Energy Systems program (DISPATCHES), and is
# copyright (c) 2020-2023 by the software owners: The Regents of the University
# of California, through Lawrence Berkeley National Laboratory, National
# Technology & Engineering Solutions of Sandia, LLC, Alliance for Sustainable
# Energy, LLC, Battelle Energy Alliance, LLC, University of Notre Dame du Lac, et
# al. All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. Both files are also available online at the URL:
# "https://github.com/gmlc-dispatches/dispatches".
#################################################################################

# Pyomo imports
from pyomo.environ import (
    Var,
    RangeSet,
    Constraint,
    NonNegativeReals,
    Reference,
    TransformationFactory,
    units as pyunits,
    sqrt,
    log,
)
from pyomo.common.config import ConfigValue, In
from pyomo.network import Arc

# IDAES imports
from idaes.core import (declare_process_block_class,
                        UnitModelBlockData)
from idaes.models.unit_models import Heater
from idaes.core.util.config import is_physical_parameter_block
from idaes.core.util.initialization import propagate_state
from idaes.core.util.constants import Constants
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.core.util import from_json, to_json, StoreSpec
from idaes.core.solvers import get_solver
import idaes.logger as idaeslog
import idaes.core.util.scaling as iscale

# Set up logger
_log = idaeslog.getLogger(__name__)


def u_tes(r, k, a, b):
    zz = r + ((a ** 3 * (4 * b ** 2 - a ** 2) + a * b ** 4 * (4 * log(b / a) - 3)) /
              (4 * k * (b ** 2 - a ** 2) ** 2))
    return 1 / zz


def _add_inlet_outlet_ports(self):
    operating_mode = self.config.operating_mode

    if operating_mode == "charge" or operating_mode == "combined":
        self.charge_inlet = self.config.property_package.build_state_block(
            self.flowsheet().time,
            defined_state=True,
            has_phase_equilibrium=True,
        )

        self.charge_outlet = self.config.property_package.build_state_block(
            self.flowsheet().time,
            defined_state=True,
            has_phase_equilibrium=True,
        )

        @self.Constraint(self.flowsheet().time, self.time_periods)
        def charge_inlet_flow_mol_equality(blk, t, p):
            return (
                blk.charge_inlet[t].flow_mol == 
                blk.num_tubes * blk.period[p].tube_charge.inlet.flow_mol[t]
            )

        @self.Constraint(self.flowsheet().time, self.time_periods)
        def charge_inlet_enth_mol_equality(blk, t, p):
            return (
                blk.charge_inlet[t].enth_mol == blk.period[p].tube_charge.inlet.enth_mol[t]
            )

        @self.Constraint(self.flowsheet().time, self.time_periods)
        def charge_inlet_pressure_equality(blk, t, p):
            return (
                blk.charge_inlet[t].pressure == blk.period[p].tube_charge.inlet.pressure[t]
            )

        p = len(self.time_periods)

        @self.Constraint(self.flowsheet().time)
        def charge_outlet_flow_mol_equality(blk, t):
            return (
                blk.charge_outlet[t].flow_mol == 
                blk.num_tubes * blk.period[p].tube_charge.outlet.flow_mol[t]
            )

        @self.Constraint(self.flowsheet().time)
        def charge_outlet_enth_mol_equality(blk, t):
            return (
                blk.charge_outlet[t].enth_mol == blk.period[p].tube_charge.outlet.enth_mol[t]
            )

        @self.Constraint(self.flowsheet().time)
        def charge_outlet_pressure_equality(blk, t):
            return (
                blk.charge_outlet[t].pressure == blk.period[p].tube_charge.outlet.pressure[t]
            )

        self.add_port(name="inlet_charge", block=self.charge_inlet,)
        self.add_port(name="outlet_charge", block=self.charge_outlet,)

    if operating_mode == "discharge" or operating_mode == "combined":
        self.discharge_inlet = self.config.property_package.build_state_block(
            self.flowsheet().time,
            defined_state=True,
            has_phase_equilibrium=True,
        )

        self.discharge_outlet = self.config.property_package.build_state_block(
            self.flowsheet().time,
            defined_state=True,
            has_phase_equilibrium=True,
        )

        @self.Constraint(self.flowsheet().time, self.time_periods)
        def discharge_inlet_flow_mol_equality(blk, t, p):
            return (
                blk.discharge_inlet[t].flow_mol == blk.num_tubes *
                blk.period[p].tube_discharge.inlet.flow_mol[t]
            )

        @self.Constraint(self.flowsheet().time, self.time_periods)
        def discharge_inlet_enth_mol_equality(blk, t, p):
            return (
                blk.discharge_inlet[t].enth_mol == blk.period[p].tube_discharge.inlet.enth_mol[t]
            )

        @self.Constraint(self.flowsheet().time, self.time_periods)
        def discharge_inlet_pressure_equality(blk, t, p):
            return (
                blk.discharge_inlet[t].pressure == blk.period[p].tube_discharge.inlet.pressure[t]
            )

        p = len(self.time_periods)

        @self.Constraint(self.flowsheet().time)
        def discharge_outlet_flow_mol_equality(blk, t):
            return (
                blk.discharge_outlet[t].flow_mol == blk.num_tubes *
                blk.period[p].tube_discharge.outlet.flow_mol[t]
            )

        @self.Constraint(self.flowsheet().time)
        def discharge_outlet_enth_mol_equality(blk, t):
            return (
                blk.discharge_outlet[t].enth_mol ==
                blk.period[p].tube_discharge.outlet.enth_mol[t]
            )

        @self.Constraint(self.flowsheet().time)
        def discharge_outlet_pressure_equality(blk, t):
            return (
                blk.discharge_outlet[t].pressure ==
                blk.period[p].tube_discharge.outlet.pressure[t]
            )

        self.add_port(name="inlet_discharge", block=self.discharge_inlet,)
        self.add_port(name="outlet_discharge", block=self.discharge_outlet,)


@declare_process_block_class("ConcreteBlock")
class ConcreteBlockData(UnitModelBlockData):
    """
    Concrete block Model Class.
    """
    CONFIG = UnitModelBlockData.CONFIG()
    CONFIG.declare(
        "segments_set",
        ConfigValue(
            doc="""Pointer to the set of segments""",
        ),
    )

    # noinspection PyAttributeOutsideInit
    def build(self):
        super().build()

        # Declare variables
        self.therm_cond = Var(
            within=NonNegativeReals,
            bounds=(0, 10),
            initialize=1,
            doc="Thermal conductivity of the concrete",
            units=pyunits.W / (pyunits.m * pyunits.K),
        )
        self.delta_time = Var(
            within=NonNegativeReals,
            bounds=(0, 4000),
            initialize=3600,
            doc="Delta for discretization of time",
            units=pyunits.s,
        )
        self.dens_mass = Var(
            within=NonNegativeReals,
            bounds=(0, 3000),
            initialize=2240,
            doc="Concrete density",
            units=pyunits.kg / pyunits.m ** 3,
        )
        self.cp_mass = Var(
            within=NonNegativeReals,
            bounds=(0, 1000),
            initialize=900,
            doc="Concrete specific heat",
            units=pyunits.J / (pyunits.kg * pyunits.K),
        )
        self.face_area = Var(
            within=NonNegativeReals,
            bounds=(0.003, 0.015),
            initialize=0.01,
            doc='Face (cross-sectional) area of the concrete wall',
            units=pyunits.m ** 2,
        )
        self.delta_z = Var(
            within=NonNegativeReals,
            bounds=(0, 100),
            initialize=5,
            doc='Length of each concrete segment',
            units=pyunits.m,
        )

        temperature_wall_index = self.config.segments_set
        self.init_temperature = Var(
            temperature_wall_index,
            within=NonNegativeReals,
            bounds=(300, 900),
            initialize=600,
            doc='Initial concrete wall temperature profile',
            units=pyunits.K,
            )
        self.temperature = Var(
            temperature_wall_index,
            within=NonNegativeReals,
            bounds=(300, 900),
            initialize=600,
            doc='Final concrete wall temperature',
            units=pyunits.K,
        )
        self.heat_rate = Var(
            temperature_wall_index,
            bounds=(-10000, 10000),
            initialize=600,
            doc='Heat transferred from/to the concrete segment',
            units=pyunits.W,
        )

        # Declare constraints
        @self.Constraint(temperature_wall_index)
        def temp_segment_constraint(b, s):
            volume = b.face_area * b.delta_z

            return (
                b.temperature[s] == b.init_temperature[s] +
                (b.delta_time * b.heat_rate[s]) / (b.dens_mass * b.cp_mass * volume)
            )

    def initialize_build(self, state_args=None, outlvl=idaeslog.NOTSET, solver=None, optarg=None):

        init_log = idaeslog.getInitLogger(self.name, outlvl, tag="unit")
        solve_log = idaeslog.getSolveLogger(self.name, outlvl, tag="unit")
        solver = get_solver(solver=solver, options=optarg)

        # TODO: Need to print a proper error message
        assert degrees_of_freedom(self) == 0

        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = solver.solve(self, tee=slc.tee)

        init_log.info_high("Initialization Step: {}.".format(idaeslog.condition(res)))
        init_log.info("Initialization Complete.")

    def calculate_scaling_factors(self):
        super().calculate_scaling_factors()


@declare_process_block_class("TubeSideHex")
class TubeSideHexData(UnitModelBlockData):
    """ConcreteTubeSide 1D Unit Model Class.
    """

    CONFIG = UnitModelBlockData.CONFIG()
    CONFIG.declare(
        "has_pressure_change",
        ConfigValue(
            default=False,
            domain=In([True, False]),
            description="Pressure change term construction flag",
            doc="""Indicates whether terms for pressure change should be constructed,
            **default** - False.
            **Valid values:** {
            **True** - include pressure change terms,
            **False** - exclude pressure change terms.}""",
        ),
    )
    CONFIG.declare(
        "property_package",
        ConfigValue(
            default=None,
            domain=is_physical_parameter_block,
            description="Property package to use for control volume",
            doc="""Property parameter object used to define property calculations 
            (default = 'use_parent_value')
            - 'use_parent_value' - get package from parent (default = None)
            - a ParameterBlock object""",
        ),
    )
    CONFIG.declare(
        "property_package_args",
        ConfigValue(
            default={},
            description="Arguments for constructing shell property package",
            doc="""A dict of arguments to be passed to the PropertyBlockData
            and used when constructing these (default = 'use_parent_value')
            - 'use_parent_value' - get package from parent (default = None)
            - a dict (see property package for documentation)""",
        ),
    )
    CONFIG.declare(
        "operating_mode",
        ConfigValue(
            default="charge",
            domain=In(["charge", "discharge"]),
            doc="""Mode of operation. """,
        ),
    )
    CONFIG.declare(
        "segments_set",
        ConfigValue(
            doc="""Pointer to the set of segments""",
        ),
    )
    CONFIG.declare(
        "use_surrogate",
        ConfigValue(
            default=False,
            domain=In([True, False]),
            description="Surrogate construction flag",
            doc="""Indicates whether a surrogate model should be used
            for enthalpy-temperature relation""",
        ),
    )
    CONFIG.declare(
        "surrogate_data",
        ConfigValue(
            default=[],
            doc="""Pointer to the points needed to construct the surrogate""",
        ),
    )

    # noinspection PyAttributeOutsideInit
    def build(self):

        # Call UnitModel.build to setup dynamics
        super().build()

        segments = self.config.segments_set
        op_mode = self.config.operating_mode

        self.hex = Heater(
            segments,
            property_package=self.config.property_package,
            has_pressure_change=self.config.has_pressure_change,
        )

        # Add the tube diameter variable
        self.tube_outer_dia = Var(
            domain=NonNegativeReals,
            initialize=0.01,
            doc="Diameter of the outer surface of the tube",
            units=pyunits.m,
        )

        # Add the tube_length variable
        self.tube_length = Var(
            domain=NonNegativeReals,
            initialize=50,
            doc="Inner surface area of the tube segment",
            units=pyunits.m,
        )

        # Connect heat exchangers with arcs
        if op_mode == "charge":
            for i in range(1, len(segments)):
                setattr(self, "sc" + str(i), Arc(source=self.hex[i].outlet,
                                                 destination=self.hex[i + 1].inlet))
                
        elif op_mode == "discharge":
            for i in range(1, len(segments)):
                setattr(self, "sc" + str(i), Arc(source=self.hex[i + 1].outlet,
                                                 destination=self.hex[i].inlet))

        tube_units = self.config.property_package.get_metadata().get_derived_units
        # Add the additional variables and heat transfer constraint
        for i in segments:
            obj = self.hex[i]

            # Add a variable for tube diameter, segment length, heat transfer coefficient 
            # and the temperature of the wall
            obj.tube_outer_dia = Var(
                domain=NonNegativeReals,
                initialize=0.01,
                doc="Outer diameter of the tube",
                units=tube_units("length"),
            )
            obj.segment_length = Var(
                domain=NonNegativeReals,
                initialize=4,
                doc="Length of each segment",
                units=tube_units("length"),
            )
            obj.htc = Var(
                domain=NonNegativeReals,
                initialize=70,
                doc="Heat transfer coefficient",
                units=tube_units("heat_transfer_coefficient"),
            )
            obj.temperature_wall = Var(
                domain=NonNegativeReals,
                bounds=(300, 900),
                initialize=500,
                doc="Temperature of the concrete block",
                units=tube_units("temperature"),
            )
            
            if self.config.use_surrogate:
                pass
            else:
                # Add energy transfer between tube wall and tube
                obj.tube_heat_transfer_eq = Constraint(
                    expr=obj.heat_duty[0] == obj.htc
                         * (Constants.pi * obj.tube_outer_dia * obj.segment_length)
                         * (obj.temperature_wall -
                            obj.control_volume.properties_out[0].temperature),
                    doc="Convective heat transfer",
                )

        # Add constraints to equate the tube diameter and length
        @self.Constraint(segments)
        def equate_tube_diameter(blk, s):
            return blk.tube_outer_dia == blk.hex[s].tube_outer_dia

        @self.Constraint(segments)
        def equate_tube_length(blk, s):
            return blk.tube_length == blk.hex[s].segment_length * len(segments)

        TransformationFactory("network.expand_arcs").apply_to(self)

    @property
    def inlet(self):
        if self.config.operating_mode == "charge":
            return self.hex[1].inlet
        elif self.config.operating_mode == "discharge":
            return self.hex[len(self.config.segments_set)].inlet

    @property
    def outlet(self):
        if self.config.operating_mode == "charge":
            return self.hex[len(self.config.segments_set)].outlet
        elif self.config.operating_mode == "discharge":
            return self.hex[1].outlet

    def initialize_build(
        self, state_args=None, outlvl=idaeslog.NOTSET, solver=None, optarg=None
    ):
        """
        Initialization routine for the unit (default solver ipopt).

        Keyword Arguments:
            state_args : a dict of arguments to be passed to the property
                         package(s) to provide an initial state for
                         initialization (see documentation of the specific
                         property package) (default = {}).
            outlvl : sets output level of initialization routine
            optarg : solver options dictionary object (default={'tol': 1e-6})
            solver : str indicating whcih solver to use during
                     initialization (default = 'ipopt')

        Returns:
            None
        """
        init_log = idaeslog.getInitLogger(self.name, outlvl, tag="unit")
        solve_log = idaeslog.getSolveLogger(self.name, outlvl, tag="unit")

        solver = get_solver(solver=solver, options=optarg)
        segment_length = self.tube_length.value / len(self.config.segments_set)

        if self.config.operating_mode == "charge":
            for i in self.config.segments_set:
                self.hex[i].tube_outer_dia.fix(self.tube_outer_dia.value)
                self.hex[i].segment_length.fix(segment_length)

                self.hex[i].initialize(outlvl=outlvl)

                self.hex[i].tube_outer_dia.unfix()
                self.hex[i].segment_length.unfix()

                if i != len(self.config.segments_set):
                    propagate_state(eval("self.sc" + str(i)))

        elif self.config.operating_mode == "discharge":
            for i in range(len(self.config.segments_set), 0, -1):
                self.hex[i].tube_outer_dia.fix(self.tube_outer_dia.value)
                self.hex[i].segment_length.fix(self.tube_length)

                self.hex[i].initialize(outlvl=outlvl)

                self.hex[i].tube_outer_dia.unfix()
                self.hex[i].segment_length.unfix()

                if i != 1:
                    propagate_state(eval("self.sc" + str(i - 1)))

        init_log.info_high("Initialization Step 1 Complete.")

        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = solver.solve(self, tee=slc.tee)
        init_log.info_high("Initialization Step 2 {}.".format(idaeslog.condition(res)))

        init_log.info("Initialization Complete.")

    def calculate_scaling_factors(self):
        super().calculate_scaling_factors()

        # for i in self.config.segments_set:
        #     iscale.constraint_scaling_transform(
        #         self.hex[i].tube_heat_transfer_eq, 1)


@declare_process_block_class("ConcreteTES")
class ConcreteTESData(UnitModelBlockData):
    """
    Concrete TES Model Class.
    """

    CONFIG = UnitModelBlockData.CONFIG()

    CONFIG.declare(
        "model_data",
        ConfigValue(
            default=None,
            description="Pointer to the dictionary containing model data"
        ),
    )
    CONFIG.declare(
        "has_pressure_change",
        ConfigValue(
            default=False,
            domain=In([True, False]),
            description="Pressure change term construction flag",
            doc="""Indicates whether terms for pressure change should be constructed,
                **default** - False.
                **Valid values:** {
                **True** - include pressure change terms,
                **False** - exclude pressure change terms.}""",
        ),
    )
    CONFIG.declare(
        "property_package",
        ConfigValue(
            default=None,
            domain=is_physical_parameter_block,
            description="Property package to use for control volume",
            doc="""Property parameter object used to define property calculations 
                (default = 'use_parent_value')
                - 'use_parent_value' - get package from parent (default = None)
                - a ParameterBlock object""",
        ),
    )
    CONFIG.declare(
        "property_package_args",
        ConfigValue(
            default={},
            description="Arguments for constructing shell property package",
            doc="""A dict of arguments to be passed to the PropertyBlockData
                and used when constructing these (default = 'use_parent_value')
                - 'use_parent_value' - get package from parent (default = None)
                - a dict (see property package for documentation)""",
        ),
    )
    CONFIG.declare(
        "operating_mode",
        ConfigValue(
            default="charge",
            domain=In(["charge", "discharge", "combined"]),
            doc="""Mode of operation. """,
        ),
    )
    CONFIG.declare(
        "use_surrogate",
        ConfigValue(
            default=False,
            domain=In([True, False]),
            description="Surrogate construction flag",
            doc="""Indicates whether a surrogate model should be used
                for enthalpy-temperature relation""",
        ),
    )
    CONFIG.declare(
        "inlet_pressure_data",
        ConfigValue(
            default={"charge": 24235081, "discharge": 8.5e5},
            doc="""Pointer to the points needed to construct the surrogate""",
        ),
    )

    # noinspection PyAttributeOutsideInit
    def build(self):
        super().build()

        data = self.config.model_data
        required_prop_list = [
            "num_tubes", "num_segments", "num_time_periods", "tube_length", 
            "tube_diameter", "therm_cond_concrete", "dens_mass_concrete", 
            "cp_mass_concrete", "init_temperature_concrete", "face_area", 
        ]
        for prop in required_prop_list:
            if prop not in data:
                raise KeyError(f"Property {prop}, a required argument, is not in model_data")

        data["delta_time"] = 3600 / data["num_time_periods"]

        property_package = self.config.property_package
        operating_mode = self.config.operating_mode

        self.time_periods = RangeSet(data['num_time_periods'])
        self.num_tubes = Var(
            within=NonNegativeReals,
            bounds=(1, 100000),
            doc='Number of tubes of the concrete TES',
            units=None,
        )
        self.num_tubes.fix(data['num_tubes'])
        self.segments = RangeSet(data["num_segments"])

        # Construct the model for each time period
        @self.Block(self.time_periods)
        def period(m):
            # Add the concrete side model
            m.concrete = ConcreteBlock(
                segments_set=m.parent_block().segments,
            )

            # Add the tube side charge model
            if operating_mode == "charge" or operating_mode == "combined":
                m.tube_charge = TubeSideHex(
                    property_package=property_package,
                    has_pressure_change=self.config.has_pressure_change,
                    segments_set=m.parent_block().segments,
                    operating_mode="charge",
                )

                @m.Constraint(m.parent_block().segments)
                def temperature_equality_constraints_charge(blk, s):
                    return blk.tube_charge.hex[s].temperature_wall == blk.concrete.temperature[s]

            # Add the tube side discharge model
            if operating_mode == "discharge" or operating_mode == "combined":
                m.tube_discharge = TubeSideHex(
                    property_package=property_package,
                    has_pressure_change=self.config.has_pressure_change,
                    segments_set=m.parent_block().segments,
                    operating_mode="discharge",
                )

                @m.Constraint(m.parent_block().segments)
                def temperature_equality_constraints_discharge(blk, s):
                    return blk.tube_discharge.hex[s].temperature_wall == blk.concrete.temperature[s]

            # Linking constraint for the net Q_fluid var in the concrete model
            # Q_fluid = Q_charge - Q_discharge
            @m.Constraint(m.parent_block().segments)
            def heat_balance_constraints(blk, s):
                if operating_mode == "charge":
                    return blk.concrete.heat_rate[s] == -blk.tube_charge.hex[s].heat_duty[0]
                elif operating_mode == "discharge":
                    return blk.concrete.heat_rate[s] == -blk.tube_discharge.hex[s].heat_duty[0]
                elif operating_mode == "combined":
                    return (blk.concrete.heat_rate[s] == -blk.tube_charge.hex[s].heat_duty[0]
                            - blk.tube_discharge.hex[s].heat_duty[0])

            return m

        _add_inlet_outlet_ports(self)

        # Add constraints connecting different time periods
        @self.Constraint(self.time_periods, self.segments)
        def initial_temperature_constraints(blk, p, s):
            if p == 1:
                return Constraint.Skip
            return blk.period[p].concrete.init_temperature[s] == blk.period[p - 1].concrete.temperature[s]

        # Add the heat transfer coefficient surrogate model
        def htc_surrogate():
            # TODO: Need to declare a variable for outer radius of the
            #       concrete block and do the following calculation in a constraint
            # Also, do this inside each hex model instead of doing it
            # outside. Reduces a lot of additional constraints!
            face_area = data["face_area"]
            a = data["tube_diameter"] / 2  # Inner radius of the concrete block
            b = sqrt(face_area / Constants.pi + a ** 2)  # Outer radius
            k_red = 0.8  # Reduction factor for thermal conductivity
            k = data["therm_cond_concrete"] * k_red
            correction_factor = 1.31

            # Tube calculations are a bit complex. If inner diameter of the
            # tube is available, then r = log(OD/ID) * (OD/2) / k_tube. Since
            # ID is not readily available, we are setting r = 0.0001 (works very well)
            return u_tes(r=0.0001, k=k, a=a, b=b) / correction_factor

        # Fix the geometry of the tube, tube inlet conditions, and material properties of concrete
        for p in self.time_periods:
            self.period[p].concrete.delta_time.fix(data["delta_time"])
            self.period[p].concrete.therm_cond.fix(data["therm_cond_concrete"])
            self.period[p].concrete.dens_mass.fix(data["dens_mass_concrete"])
            self.period[p].concrete.cp_mass.fix(data["cp_mass_concrete"])
            self.period[p].concrete.face_area.fix(data["face_area"])
            self.period[p].concrete.delta_z.fix(data["tube_length"] / data["num_segments"])

            if operating_mode == "charge" or operating_mode == "combined":
                self.period[p].tube_charge.tube_outer_dia.fix(data["tube_diameter"])
                self.period[p].tube_charge.tube_length.fix(data["tube_length"])

                htc = htc_surrogate()
                for s in self.segments:
                    self.period[p].tube_charge.hex[s].htc.fix(htc)

            if operating_mode == "discharge" or operating_mode == "combined":
                self.period[p].tube_discharge.tube_outer_dia.fix(data["tube_diameter"])
                self.period[p].tube_discharge.tube_length.fix(data["tube_length"])

                htc = htc_surrogate()
                for s in self.segments:
                    self.period[p].tube_discharge.hex[s].htc.fix(htc)

        # Fixing the initial concrete temperature for the first time block
        for s in self.segments:
            self.period[1].concrete.init_temperature[s].fix(data["init_temperature_concrete"][s - 1])

    def initialize_build(self, state_args=None, outlvl=idaeslog.INFO_HIGH, solver=None, optarg=None):
        """
        Initialization routine for the entire model (default solver ipopt).

        Keyword Arguments:
            state_args : a dict of arguments to be passed to the property
                         package(s) to provide an initial state for
                         initialization (see documentation of the specific
                         property package) (default = {}).
            outlvl : sets output level of initialization routine
            optarg : solver options dictionary object (default={'tol': 1e-6})
            solver : str indicating whcih solver to use during
                     initialization (default = 'ipopt')

        Returns:
            None
        """
        init_log = idaeslog.getInitLogger(self.name, outlvl, tag="unit")
        solve_log = idaeslog.getSolveLogger(self.name, outlvl, tag="unit")

        # Store original specification so initialization doesn't change the model
        # This will only restore the values of variables that were originally fixed
        sp = StoreSpec.value_isfixed_isactive(only_fixed=True)

        if hasattr(self, "inlet_charge"):
            state_charge = to_json(self.charge_inlet, return_dict=True, wts=sp)
            self.inlet_charge.fix()

        if hasattr(self, "inlet_discharge"):
            state_discharge = to_json(self.discharge_inlet, return_dict=True, wts=sp)
            self.inlet_discharge.fix()

        operating_mode = self.config.operating_mode
        flags_charge = {}
        flags_discharge = {}
        if solver is None:
            solver = get_solver()

        solver.options = {
            # "tol": 1e-6,
            "max_iter": 100,
            # "halt_on_ampl_error": "yes",
            "bound_push": 1e-6,
            # "mu_init": 1e-5
        }

        for p in self.time_periods:
            if operating_mode == "charge" or operating_mode == "combined":
                tube_inlet = self.period[p].tube_charge.inlet
                tube_inlet.flow_mol[0].value = (self.inlet_charge.flow_mol[0].value /
                                                self.num_tubes.value)
                tube_inlet.pressure[0].value = self.inlet_charge.pressure[0].value
                tube_inlet.enth_mol[0].value = self.inlet_charge.enth_mol[0].value

            if operating_mode == "discharge" or operating_mode == "combined":
                tube_inlet = self.period[p].tube_discharge.inlet
                tube_inlet.flow_mol[0].value = (self.inlet_discharge.flow_mol[0].value /
                                                self.num_tubes.value)
                tube_inlet.pressure[0].value = self.inlet_discharge.pressure[0].value
                tube_inlet.enth_mol[0].value = self.inlet_discharge.enth_mol[0].value

        for p in self.time_periods:
            # Fix the initial temperature for p > 1
            if p > 1:
                for t in self.segments:
                    self.period[p].concrete.init_temperature[t].fix(
                        self.period[p - 1].concrete.temperature[t].value)

            # **************************************************
            #               INITIALIZING TUBE SIDE
            # **************************************************
            # TODO: Need to print a proper error message
            if operating_mode == "charge" or operating_mode == "combined":
                # Use the initial concrete block temperature for flux calculation
                for s in self.segments:
                    self.period[p].tube_charge.hex[s].temperature_wall.fix(
                        self.period[p].concrete.init_temperature[s].value)

                # Hold the state of the inlet
                flags_charge[p] = self.period[p].tube_charge.hex[1].control_volume.initialize(hold_state=True)
                self.period[p].tube_charge.initialize(outlvl=outlvl, optarg=solver.options)

                # Unfix the final wall temperature
                for s in self.segments:
                    self.period[p].tube_charge.hex[s].temperature_wall.unfix()

            if operating_mode == "discharge" or operating_mode == "combined":
                # Use the initial concrete block temperature for flux calculation
                for s in self.segments:
                    self.period[p].tube_discharge.hex[s].temperature_wall.fix(
                        self.period[p].concrete.init_temperature[s].value)

                # Hold the state of the inlet
                flags_discharge[p] = self.period[p].tube_discharge.hex[len(self.segments)].\
                    control_volume.initialize(hold_state=True)
                self.period[p].tube_discharge.initialize(outlvl=outlvl, optarg=solver.options)

                # Unfix the final wall temperature
                for s in self.segments:
                    self.period[p].tube_discharge.hex[s].temperature_wall.unfix()

            # **************************************************
            #            INITIALIZING CONCRETE SIDE
            # **************************************************
            for s in self.segments:
                if operating_mode == "charge":
                    heat_rate = -self.period[p].tube_charge.hex[s].heat_duty[0].value
                elif operating_mode == "discharge":
                    heat_rate = -self.period[p].tube_discharge.hex[s].heat_duty[0].value
                else:
                    heat_rate = - self.period[p].tube_charge.hex[s].heat_duty[0].value \
                                - self.period[p].tube_discharge.hex[s].heat_duty[0].value

                self.period[p].concrete.heat_rate[s].fix(heat_rate)

            self.period[p].concrete.initialize(outlvl=outlvl, optarg=solver.options)

            # Unfix q_fluid
            self.period[p].concrete.heat_rate.unfix()

            # **************************************************
            #               INITIALIZING TUBE + CONCRETE SIDES
            # **************************************************
            assert degrees_of_freedom(self.period[p]) == 0
            with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
                res = solver.solve(self.period[p], tee=slc.tee)
            init_log.info_high("Initialization of tube and concrete: {}.".format(idaeslog.condition(res)))

            # Unfix the initial concrete temperature for p > 1
            if p > 1:
                self.period[p].concrete.init_temperature.unfix()

        # **************************************************
        #               INITIALIZING THE ENTIRE MODEL
        # **************************************************
        p = len(self.time_periods)

        if operating_mode == "charge" or operating_mode == "combined":
            tube_outlet = self.period[p].tube_charge.outlet
            self.outlet_charge.flow_mol[0].value = \
                (tube_outlet.flow_mol[0].value * self.num_tubes.value)
            self.outlet_charge.pressure[0].value = tube_outlet.pressure[0].value
            self.outlet_charge.enth_mol[0].value = tube_outlet.enth_mol[0].value

        if operating_mode == "discharge" or operating_mode == "combined":
            tube_outlet = self.period[p].tube_discharge.outlet
            self.outlet_discharge.flow_mol[0].value = \
                (tube_outlet.flow_mol[0].value * self.num_tubes.value)
            self.outlet_discharge.pressure[0].value = tube_outlet.pressure[0].value
            self.outlet_discharge.enth_mol[0].value = tube_outlet.enth_mol[0].value

        for p in self.time_periods:
            if operating_mode == "charge" or operating_mode == "combined":
                self.period[p].tube_charge.hex[1].control_volume.release_state(flags_charge[p])

            if operating_mode == "discharge" or operating_mode == "combined":
                self.period[p].tube_discharge.hex[len(self.segments)].control_volume.release_state(flags_discharge[p])

        assert degrees_of_freedom(self) == 0
        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = solver.solve(self, tee=slc.tee)

        if hasattr(self, "inlet_charge"):
            from_json(self.charge_inlet, sd=state_charge, wts=sp)
        if hasattr(self, "inlet_discharge"):
            from_json(self.discharge_inlet, sd=state_discharge, wts=sp)

        init_log.info_high("Initialization TES: {}.".format(idaeslog.condition(res)))
        init_log.info("Initialization Complete.")

    def set_default_scaling_factors(self):
        for p in self.time_periods:
            tube_charge = self.period[p].tube_charge
            tube_discharge = self.period[p].tube_discharge
            concrete = self.period[p].concrete

            # iscale.set_scaling_factor(concrete.q_fluid, 1e-3)
            # iscale.set_scaling_factor(concrete.temperature, 1e-2)

            # if p != 1:
            #     iscale.set_scaling_factor(concrete.init_temperature, 1e-2)

            for i in self.segments:
                # Charge model
                iscale.set_scaling_factor(
                    tube_charge.hex[i].control_volume.heat[0], 1e-3)
                iscale.set_scaling_factor(
                    tube_charge.hex[i].inlet.flow_mol[0], 1)
                iscale.set_scaling_factor(
                    tube_charge.hex[i].outlet.flow_mol[0], 1)
                # iscale.set_scaling_factor(
                #     tube_charge.hex[i].d_tube_outer, 1e2)
                # iscale.set_scaling_factor(
                #     tube_charge.hex[i].tube_length, 1e-1)
                # iscale.set_scaling_factor(
                #     tube_charge.hex[i].htc, 1e-2)
                # iscale.set_scaling_factor(
                #     tube_charge.hex[i].temperature_wall, 1e-2)

                # Discharge model
                iscale.set_scaling_factor(
                    tube_discharge.hex[i].control_volume.heat[0], 1e-3)
                iscale.set_scaling_factor(
                    tube_discharge.hex[i].inlet.flow_mol[0], 1)
                iscale.set_scaling_factor(
                    tube_discharge.hex[i].outlet.flow_mol[0], 1)
                # iscale.set_scaling_factor(
                #     tube_discharge.hex[i].d_tube_outer, 1e2)
                # iscale.set_scaling_factor(
                #     tube_discharge.hex[i].tube_length, 1e-1)
                # iscale.set_scaling_factor(
                #     tube_discharge.hex[i].htc, 1e-2)
                # iscale.set_scaling_factor(
                #     tube_discharge.hex[i].temperature_wall, 1e-2)




#################################################################################
# DISPATCHES was produced under the DOE Design Integration and Synthesis Platform
# to Advance Tightly Coupled Hybrid Energy Systems program (DISPATCHES), and is
# copyright (c) 2020-2023 by the software owners: The Regents of the University
# of California, through Lawrence Berkeley National Laboratory, National
# Technology & Engineering Solutions of Sandia, LLC, Alliance for Sustainable
# Energy, LLC, Battelle Energy Alliance, LLC, University of Notre Dame du Lac, et
# al. All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. Both files are also available online at the URL:
# "https://github.com/gmlc-dispatches/dispatches".
#################################################################################
"""
Tests for Thermal Oil property package.
Authors: Naresh Susarla
"""

import pytest
from pyomo.environ import (ConcreteModel,
                           Constraint,
                           Expression,
                           Param,
                           value,
                           Var)
from pyomo.util.check_units import assert_units_consistent
from idaes.core import (MaterialBalanceType,
                        EnergyBalanceType)
import idaes.logger as idaeslog
from dispatches.properties import thermaloil_properties


@pytest.fixture(scope="class")
def model():
    model = ConcreteModel()
    model.params = thermaloil_properties.ThermalOilParameterBlock()
    model.props = model.params.build_state_block([1])
    return model


@pytest.mark.unit
def test_config(model):
    assert len(model.params.config) == 1


@pytest.mark.unit
def test_build(model):

    assert len(model.params.phase_list) == 1
    for i in model.params.phase_list:
        assert i == "Liq"

    assert len(model.params.component_list) == 1
    for i in model.params.component_list:
        assert i in ['therminol66']

    assert isinstance(model.params.cp_param_1, Param)
    assert value(model.params.cp_param_1) == 1496.005

    assert isinstance(model.params.cp_param_2, Param)
    assert value(model.params.cp_param_2) == 3.313

    assert isinstance(model.params.cp_param_3, Param)
    assert value(model.params.cp_param_3) == 0.0008970785

    assert isinstance(model.params.rho_param_1, Param)
    assert value(model.params.rho_param_1) == 1026.7

    assert isinstance(model.params.rho_param_2, Param)
    assert value(model.params.rho_param_2) == -0.7281

    assert isinstance(model.params.nu_param_1, Param)
    assert value(model.params.nu_param_1) == 586.375

    assert isinstance(model.params.nu_param_2, Param)
    assert value(model.params.nu_param_2) == 62.5

    assert isinstance(model.params.nu_param_3, Param)
    assert value(model.params.nu_param_3) == -2.2809

    assert isinstance(model.params.nu_param_4, Param)
    assert value(model.params.nu_param_4) == 1E-6

    assert isinstance(model.params.kappa_param_1, Param)
    assert value(model.params.kappa_param_1) == 0.118294

    assert isinstance(model.params.kappa_param_2, Param)
    assert value(model.params.kappa_param_2) == -3.3E-5

    assert isinstance(model.params.kappa_param_3, Param)
    assert value(model.params.kappa_param_3) == -1.5E-7

    assert isinstance(model.params.ref_temperature, Param)
    assert value(model.params.ref_temperature) == 273.15

    assert isinstance(model.props[1].flow_mass, Var)
    assert value(model.props[1].flow_mass) == 100

    assert isinstance(model.props[1].pressure, Var)
    assert value(model.props[1].pressure) == 1.01325E5

    assert isinstance(model.props[1].temperature, Var)
    assert value(model.props[1].temperature) == 523

    assert isinstance(model.props[1].enth_mass, Var)
    assert len(model.props[1].enth_mass) == 1
    for i in model.props[1].enth_mass:
        assert value(model.props[1].enth_mass[i]) == 1

    assert isinstance(model.props[1].enthalpy_eq, Constraint)
    assert isinstance(model.props[1].cp_mass, Expression)
    assert isinstance(model.props[1].dens_mass, Expression)
    assert isinstance(model.props[1].visc_k_phase, Expression)
    assert isinstance(model.props[1].visc_d_phase, Expression)
    assert isinstance(model.props[1].therm_cond_phase, Expression)


@pytest.mark.unit
def test_get_material_flow_terms(model):
    for p in model.params.phase_list:
        for j in model.params.component_list:
            assert str(
                model.props[1].get_material_flow_terms(p, j)) == str(
                model.props[1].flow_mass)


@pytest.mark.unit
def test_get_enthalpy_flow_terms(model):
    for p in model.params.phase_list:
        assert str(model.props[1].get_enthalpy_flow_terms(p)) == str(
                model.props[1].enthalpy_flow_terms[p])


@pytest.mark.unit
def test_default_material_balance_type(model):
    assert model.props[1].default_material_balance_type() == \
        MaterialBalanceType.componentTotal


@pytest.mark.unit
def test_default_energy_balance_type(model):
    assert model.props[1].default_energy_balance_type() == \
        EnergyBalanceType.enthalpyTotal


@pytest.mark.unit
def test_define_state_vars(model):
    sv = model.props[1].define_state_vars()

    assert len(sv) == 3
    for i in sv:
        assert i in ["flow_mass",
                     "temperature",
                     "pressure"]


@pytest.mark.unit
def test_define_port_members(model):
    sv = model.props[1].define_state_vars()

    assert len(sv) == 3
    for i in sv:
        assert i in ["flow_mass",
                     "temperature",
                     "pressure"]


@pytest.mark.unit
def test_initialize(model):
    assert not model.props[1].flow_mass.fixed
    assert not model.props[1].temperature.fixed
    assert not model.props[1].pressure.fixed

    model.props.initialize(hold_state=False, outlvl=idaeslog.INFO)

    assert not model.props[1].flow_mass.fixed
    assert not model.props[1].temperature.fixed
    assert not model.props[1].pressure.fixed


@pytest.mark.unit
def test_initialize_hold(model):
    assert not model.props[1].flow_mass.fixed
    assert not model.props[1].temperature.fixed
    assert not model.props[1].pressure.fixed

    flags = model.props.initialize(hold_state=True)

    assert model.props[1].flow_mass.fixed
    assert model.props[1].temperature.fixed
    assert model.props[1].pressure.fixed

    model.props.release_state(flags, outlvl=idaeslog.INFO)

    assert not model.props[1].flow_mass.fixed
    assert not model.props[1].temperature.fixed
    assert not model.props[1].pressure.fixed


@pytest.mark.unit
def check_units(model):
    assert_units_consistent(model)



#################################################################################
# DISPATCHES was produced under the DOE Design Integration and Synthesis Platform
# to Advance Tightly Coupled Hybrid Energy Systems program (DISPATCHES), and is
# copyright (c) 2020-2023 by the software owners: The Regents of the University
# of California, through Lawrence Berkeley National Laboratory, National
# Technology & Engineering Solutions of Sandia, LLC, Alliance for Sustainable
# Energy, LLC, Battelle Energy Alliance, LLC, University of Notre Dame du Lac, et
# al. All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. Both files are also available online at the URL:
# "https://github.com/gmlc-dispatches/dispatches".
#################################################################################

__author__ = "Radhakrishna Tumbalam Gooty"

import pandas as pd
from dispatches.util import SynHist_integration


def generate_syn_realizations(
    pkl_filename, 
    set_years, 
    n_scenarios=1, 
    n_days=365, 
    result_filename="synthetic_lmps",
    write_excel=False,
):

    syn_hist = SynHist_integration(pkl_filename)

    final_lmp_data = {}
    for s in range(1, n_scenarios+1):
        final_lmp_data[s] = {}

        # Generate synthetic history
        syn_hist_dict = syn_hist.generateSyntheticHistory("price", set_years)

        # set of clusters
        cluster_set = [i for i in syn_hist_dict["LMP"][set_years[0]].keys()]

        for y in set_years:
            # Cluster map day -> cluster
            cmap = [c for d in range(n_days) for c in cluster_set
                    if d in syn_hist_dict["cluster_map"][y][c]]
            
            # Clustered LMP signals
            LMP = syn_hist_dict["LMP"][y]
            lmp_list = []

            for d in range(n_days):
                _cluster_lmps = [v for v in LMP[cmap[d]].values()]
                lmp_list.extend(_cluster_lmps)

            final_lmp_data[s][y] = lmp_list


    if write_excel:
        filename = result_filename + ".xlsx"
        df_final_lmp_data = {}

        for s in range(1, n_scenarios+1):
            df_final_lmp_data[s] = pd.DataFrame(final_lmp_data[s])

        with pd.ExcelWriter(filename) as writer:
            for s in range(1, n_scenarios+1):
                df_final_lmp_data[s].to_excel(writer, sheet_name="Scenario" + str(s))

    if n_scenarios == 1:
        return final_lmp_data[1]
    
    else:
        return final_lmp_data
